### 注意正则表达式无法表达上下文无关文法. 所以正则表达式无法提取任意层次嵌套的函数调用.

import re

## 提取的格式不完整
# text = "my_function(arg1, arg2) another_call()"
text = 'group_distinct_1h(member_id+foo(totoal_fee, 34), ip)'
pattern = r"(\w+)\((.*?)\)"  # Captures function name and arguments
match = re.search(pattern, text)

if match:
    function_name = match.group(1)
    arguments = match.group(2)
    print(f"Function Name: {function_name}, Arguments: {arguments}")

## 无法识别函数调用之外的 infix notation(中缀表达式)
# text = "outer_func(arg1, inner_func(sub_arg))"
text = 'group_distinct_1h(member_id+foo(totoal_fee,bar(a)), ip)'
pattern = r"(\w+)\(([^()]*?(?:\(([^()]*)\))?[^()]*?)\)"

match = re.search(pattern, text)
if match:
    function_name = match.group(1)
    arguments_string = match.group(2)
    print(f"Function Name: {function_name}")
    print(f"Arguments String: {arguments_string}")



import logging
import inspect
import asyncio
from pprint import pprint
from pathlib import Path
from zen_rule import ZenRule, udf, FuncArg, FuncRet

from utils import httpsession, get_state

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


@udf(
    comments="test udf foo",
    args_info=[
        FuncArg(arg_name="a", arg_type="string", defaults="", comments="var a"),
        FuncArg(arg_name="b", arg_type="string", defaults="", comments="var b"),
        FuncArg(arg_name="c", arg_type="string", defaults="", comments="var c"),
    ],
    return_info=FuncRet(field_type="string", examples="fccdjny", comments="返回值示例, 字段解释")     
)
def foo(*args, **kwargs):
    logger.info(f"{inspect.stack()[0][3]} args:{args}")
    logger.info(f"{inspect.stack()[0][3]} kwargs:{kwargs}")
    return "foo value"


@udf(
    comments="group_distinct_1m_demo function",
    args_info=[
        FuncArg(arg_name="group", arg_type="string", defaults="", comments="var group"),
        FuncArg(arg_name="distinct", arg_type="string", defaults="", comments="var distinct"),
    ],
    return_info=FuncRet(field_type="object", examples={}, comments="返回值示例, 字段解释")     
)
def group_distinct_1m_demo(*args, **kwargs):
    logger.info(f"{inspect.stack()[0][3]} args:{args}")
    logger.info(f"{inspect.stack()[0][3]} kwargs:{kwargs}")
    return {
            "function": "group_distinct_1m_demo",
            "pv": 1,
            "uv": 2,
            "gpv": 3
        }


async def test_zenrule_v3():
    """
        推荐线上生产环境使用此模式进行规则执行, 可以缓存决策对象, 提高性能.
    """
    # httpsession.set(object())  # 先设置 contextvars 再创建包含 custom_handler 的zen engine 实例, 自定义函数可以 contextvars get.
    zr = ZenRule({})
    # httpsession.set(object())
    basedir = Path(__file__).parent
    filename = basedir / "graph" / "custom_v3.json"
    key = filename

    if not zr.get_decision_cache(key):
        # 根据实际情况去加载规则图的内容.
        with open(filename, "r", encoding="utf8") as f:
            logger.warning(f"graph json: %s", filename)
            content =  f.read()
        zr.create_decision_with_cache_key(key, content)  # 将规则图缓存在键下, 这样可以只读取规则一次，解析一次，然后复用决策对象 decision
    for i in range(1):
        result = await zr.async_evaluate(key, {"input": 7, "myvar": 15})
        print("zen rule custom_v3 result:", result)
        print(f"------------------{i}------------------------")


async def test_zenrulev1_with_enginev3():
    zr = ZenRule({})
    basedir = Path(__file__).parent
    filename = basedir / "graph" / "custom_v1.json"
    key = filename

    with open(filename, "r", encoding="utf8") as f:
        logger.warning(f"graph json: %s", filename)
        content =  f.read()

    zr.create_decision_with_cache_key(key, content)  # 将规则图缓存在键下, 这样可以只读取规则一次，解析一次，然后复用决策对象 decision
    result = await zr.async_evaluate(key, {"ip": "1.2.3.4", "user": "17707115956"})
    print("zen rule custom_v1 result:", result)


if __name__ == "__main__":
    # test_zenrule_v3
    asyncio.run(test_zenrule_v3())
    asyncio.run(test_zenrulev1_with_enginev3())




#  pytest tests/zen/test_zen.py -sl
#  -s 捕获标准输出
#  -l 显示出错的测试用例中的局部变量
#  -k 指定某一个测试用例执行
#  --setup-show  展示的单元测试 SETUP 和 TEARDOWN 的细节，展示测试依赖的加载和销毁顺序

import asyncio
from pathlib import Path
from pprint import pprint
import pytest
import zen

graphjson = """
{
  "contentType": "application/vnd.gorules.decision",
  "nodes": [
    {
      "type": "inputNode",
      "content": {
        "schema": ""
      },
      "id": "d4c5fdef-96c0-4c5e-8e7b-e1e4f9797251",
      "name": "request",
      "position": {
        "x": 245,
        "y": 355
      }
    },
    {
      "type": "outputNode",
      "content": {
        "schema": ""
      },
      "id": "a2cb6433-140b-4d2c-89f2-0e4f3140dd1a",
      "name": "response",
      "position": {
        "x": 1120,
        "y": 365
      }
    },
    {
      "type": "expressionNode",
      "content": {
        "expressions": [
          {
            "id": "8bf2c094-3115-4608-9416-1c854c68cc89",
            "key": "result",
            "value": "num * 2"
          }
        ],
        "passThrough": false,
        "inputField": null,
        "outputPath": null,
        "executionMode": "single"
      },
      "id": "45bbd693-493b-415d-8ad6-0d57c9a40d3e",
      "name": "expression1",
      "position": {
        "x": 715,
        "y": 400
      }
    }
  ],
  "edges": [
    {
      "id": "84fa9a9a-e90a-4322-be6f-94d3efd84edf",
      "sourceId": "d4c5fdef-96c0-4c5e-8e7b-e1e4f9797251",
      "type": "edge",
      "targetId": "45bbd693-493b-415d-8ad6-0d57c9a40d3e"
    },
    {
      "id": "0cd7ec7d-05d9-4497-abc7-d51b2202ddb4",
      "sourceId": "45bbd693-493b-415d-8ad6-0d57c9a40d3e",
      "type": "edge",
      "targetId": "a2cb6433-140b-4d2c-89f2-0e4f3140dd1a"
    }
  ]
}
"""

graphs = {"expression_multi2.json": graphjson}

def loader(key):
    # 考虑是否把所有的规则都往文件系统上缓存一份.
    return graphs[key]

# 同步测试
def test_zen_with_loader():
    """
        engine.evaluate 会隐式地调用 loader 函数
    """
    engine = zen.ZenEngine({"loader": loader})
    input = {
        "num": 15,
        "ip": '127.0.0.1',
    }
    res_with_trace = engine.evaluate("expression_multi2.json", input, {"trace": True})  # 测试同步方法
    assert "result" in res_with_trace and "trace" in res_with_trace
    res = engine.evaluate("expression_multi2.json", input)  
    assert "result" in res

    # # todo: 修复这个错误
    # engine_without_loader = zen.ZenEngine({})
    # engine_without_loader.evaluate("expression_multi2.json", input, {"trace": True})  # 测试同步方法


# 同步测试
def test_zen_with_create_decision():
    """
        zen_expression_switch_decisionTable
        这种调用模式可以复用 decision 对象. 所以线上服务推荐使用此种模式
    """
    engine = zen.ZenEngine()
    con = loader("expression_multi2.json")
    decision = engine.create_decision(con)
    input = {
        "num": 15,
        "ip": '127.0.0.1',
        "phone": '1770711xxxx',
    }
    res_with_trace = decision.evaluate(input, {"trace": True})
    # pprint(res_with_trace)
    assert "result" in res_with_trace and "trace" in res_with_trace
    res = decision.evaluate(input)
    # pprint(res)
    assert "result" in res


@pytest.mark.asyncio
async def test_zen_with_loader_async_evaluate():
    """
        异步函数需要 pytest.mark.asyncio 装饰器. 
        以及需要导入 pytest-asyncio
        uv add --dev pytest-asyncio
    """
    engine = zen.ZenEngine({"loader": loader})
    input = {
        "num": 15,
        "ip": '127.0.0.1',
    }
    res_with_trace = await engine.async_evaluate("expression_multi2.json", input, {"trace": True})  # 测试同步方法
    assert "result" in res_with_trace and "trace" in res_with_trace
    res = await engine.async_evaluate("expression_multi2.json", input)  
    assert "result" in res


import logging
import os
from pprint import pformat
from typing import List, Dict, Any
from inspect import signature, Parameter, iscoroutinefunction
from pathlib import Path
logger = logging.getLogger(__name__)


def namespace_split(target,namespace=None) -> str:
    if namespace:
        return str(namespace)
    try:
        name_list = str(target).split(".")
        new_name_list = []
        for item in name_list:
            if "functions" == item:
                new_name_list = []
            else:
                new_name_list.append(item)
        return ".".join(new_name_list)
    except Exception as e:
        logger.error(e, exc_info=True)
        return ""


# Define classes for managing function arguments and return values
class FuncArg:
    def __init__(self, arg_name: str, arg_type: str, defaults: Any, comments: str):
        self.arg_name = arg_name
        self.arg_type = arg_type
        self.defaults = defaults
        self.comments = comments

    def to_dict(self):
        return {
            "arg_name": self.arg_name,
            "arg_type": self.arg_type,
            "defaults": self.defaults,
            "comments": self.comments,
        }

class FuncRet:
    def __init__(self, field_type: str, examples: Any, comments: str):
        self.field_type = field_type
        self.examples = examples
        self.comments = comments

    def to_dict(self):
        return {
            "field_type": self.field_type,
            "examples": self.examples,
            "comments": self.comments,
        }


# UDF Manager class to manage the registered UDFs
class UDFManager:
    # 之后再考虑是否支持 zen expression 的混用了.
    def __init__(self):
        self.functions = {}
        self.param_annotation = {
            str: "string",
            dict: "object",  # json
            list: "array",
            tuple: "array",
            int: "number",
            float: "number",
        }
        self.default_type = "Any"

    def register_function(self, func, comments:str =None, args_info: FuncArg =None, return_info: FuncRet =None, namespace:str = None):
        sig = signature(func)
        arguments = []

        # Use provided args_info or derive from function signature
        if args_info:
            for info in args_info:
                arguments.append(info)
        else:
            # logger.debug(f"sig.parameters: {sig.parameters}")
            # logger.debug(f"sig.parameters items(): {pformat(sig.parameters.items())}")
            for name, param in sig.parameters.items():
                arg_type = self.param_annotation.get(param.annotation, self.default_type)
                defaults = '' if param.default == Parameter.empty else param.default
                comments = f'{name}:{arg_type}'
                arguments.append(FuncArg(name, arg_type, defaults, comments))

            extra_arguments = [(arg.arg_name, arg.arg_type) for arg in arguments if arg.arg_name not in {"args", "kwargs"}]
            if extra_arguments:
                raise Exception("udf function only allow to contains *args, **kwargs, please use @udf args_info to describe arg names")            
        # logger.debug(f"return_info:{return_info}")
        return_values = return_info.to_dict() if return_info else {}
        # logger.debug(f"return_values:{return_values}")
        self.functions[func.__name__] = {
            "func": func,
            "name": func.__name__,
            "arglength": len(arguments),
            "arguments": [arg.to_dict() for arg in arguments],
            "return_values": return_values,
            "comments": comments or func.__doc__,
            "namespace": namespace_split(func.__module__,namespace=namespace),
        }

    def get_udf_info(self) -> List[Dict[str, Any]]:
        return [
            {
                "name": data["name"],
                "arglength": data["arglength"],
                "arguments": data["arguments"],
                "return_values": data["return_values"],
                "comments": data["comments"],
                "namespace": data["namespace"],
                "kind": data["namespace"].split(".")[-1]
            }
            for data in self.functions.values()
        ]

    def udf_info(self, name) -> List[Dict[str, Any]]:
        return self.functions[name] 

    async def __call__(self, udf_name: str, *args, **kwargs) -> Any:
        """
            v2 spec
            执行器要函数解析放到一起, 在function_call_parse中实现解析器和执行器,
            然后执行器调用真正的udf函数管理和执行器.
        """
        if udf_name in self.functions:
      
            func = self.functions[udf_name]["func"]
            if iscoroutinefunction(func):
                return await func(*args, **kwargs)
            else:
                return func(*args, **kwargs)
        else:
            raise ValueError(f"Function '{udf_name}' is not registered in UDFManager")

    async def call_udf(self, udf_name: str, *args, **kwargs) -> Any:

        if udf_name in self.functions:
      
            func = self.functions[udf_name]["func"]
            if iscoroutinefunction(func):
                return await func(*args, **kwargs)
            else:
                return func(*args, **kwargs)
        else:
            raise ValueError(f"Function '{udf_name}' is not registered in UDFManager")

# Instantiate UDF Manager
udf_manager = UDFManager()

# Decorator to register UDFs with optional custom metadata
def udf(comments=None, args_info=None, return_info=None,namespace=None):
    def decorator(func):
        udf_manager.register_function(func, comments, args_info, return_info,namespace)
        return func
    return decorator




from pathlib import Path
from pprint import pformat
import logging
import json
import os.path
import unittest
import glob
import zen

logger = logging.getLogger(__name__)
basedir = Path(__file__).parent

def loader(key):
    with open(basedir / "test-data/" / key, "r") as f:
        return f.read()

def graph_loader(key):
    with open(basedir / "test-data/graphs/" / key, "r") as f:
        return f.read()

def custom_handler(request):
    p1 = request.get_field("prop1")
    return {
        "output": {"sum": p1}
    }


# The test based on unittest module
class ZenEngine(unittest.TestCase):
    def test_decision_using_loader(self):
        engine = zen.ZenEngine({"loader": loader})
        r1 = engine.evaluate("function.json", {"input": 5})
        r2 = engine.evaluate("table.json", {"input": 2})
        r3 = engine.evaluate("table.json", {"input": 12})

        self.assertEqual(r1["result"]["output"], 10)
        self.assertEqual(r2["result"]["output"], 0)
        self.assertEqual(r3["result"]["output"], 10)

    def test_decisions_using_getDecision(self):
        engine = zen.ZenEngine({"loader": loader})

        functionDecision = engine.get_decision("function.json")
        tableDecision = engine.get_decision("table.json")

        r1 = functionDecision.evaluate({"input": 10})
        r2 = tableDecision.evaluate({"input": 5})
        r3 = tableDecision.evaluate({"input": 12})

        self.assertEqual(r1["result"]["output"], 20)
        self.assertEqual(r2["result"]["output"], 0)
        self.assertEqual(r3["result"]["output"], 10)

    def test_create_decisions_from_content(self):
        engine = zen.ZenEngine()
        with open(basedir / "test-data/function.json", "r") as f:
            functionContent = f.read()
        functionDecision = engine.create_decision(functionContent)

        r = functionDecision.evaluate({"input": 15})
        self.assertEqual(r["result"]["output"], 30)

    def test_engine_custom_handler(self):
        engine = zen.ZenEngine({"loader": loader, "customHandler": custom_handler})
        r1 = engine.evaluate("custom.json", {"a": 10})
        r2 = engine.evaluate("custom.json", {"a": 20})
        r3 = engine.evaluate("custom.json", {"a": 30})

        self.assertEqual(r1["result"]["sum"], 20)
        self.assertEqual(r2["result"]["sum"], 30)
        self.assertEqual(r3["result"]["sum"], 40)

    def test_evaluate_expression(self):
        result = zen.evaluate_expression("sum(a)", {"a": [1, 2, 3, 4]})
        self.assertEqual(result, 10)

    def test_evaluate_unary_expression(self):
        result = zen.evaluate_unary_expression("'FR', 'ES', 'GB'", {"$": "GB"})
        self.assertEqual(result, True)

    def test_render_template(self):
        result = zen.render_template("{{ a + b }}", {"a": 10, "b": 20})
        self.assertEqual(result, 30)

    def test_sleep_function(self):
        engine = zen.ZenEngine({"loader": loader, "customHandler": custom_handler})

        engine.evaluate("sleep-function.json", {})
        self.assertTrue(True)

    def test_http_function(self):
        engine = zen.ZenEngine({"loader": loader, "customHandler": custom_handler})

        engine.evaluate("http-function.json", {})
        self.assertTrue(True)

    def test_additional_options(self):
        engine = zen.ZenEngine({"loader": loader, "customHandler": custom_handler})

        engine.evaluate("sleep-function.json", {}, {"trace": True})
        self.assertTrue(True)

    def test_evaluate_graphs(self):
        engine = zen.ZenEngine({"loader": graph_loader})
        json_files = glob.glob(str(basedir / "test-data/graphs/*.json"))

        for json_file in json_files:
            logger.info(json_file)
            with open(json_file, "r") as f:
                json_contents = json.loads(f.read())

            for test_case in json_contents["tests"]:
                key = os.path.basename(json_file)

                engine_response = engine.evaluate(key, test_case["input"])
                decision = engine.get_decision(key)
                decision_response = decision.evaluate(test_case["input"])

                self.assertEqual(engine_response["result"], test_case["output"])
                self.assertEqual(decision_response["result"], test_case["output"])

if __name__ == '__main__':
    unittest.main()



import re


def op_args_combination(args, function_meta):
    """
        将参数值和参数名字组装为字典.
        如果没有在自定义函数中设置对应的 arg_name, 那么默认的关键字参数的下标就是以下划线加数字(如_0, _1, _2 )等开头.
    """
    arguments = [i.get("arg_name") for i in function_meta["arguments"] if i.get("arg_name") not in {"args", "kwargs"}]
    if not arguments:
        arguments = [f"_{i}" for i in range(len(args))]
    keyword_args = {arg_name: v for arg_name, v in zip(arguments, args)}
    return keyword_args


def parse_oprator_expr_v3(expr):
    """
        v3自定义节点中的算子 oprater 调用支持两种格式:
        1.  ;; 当作起始符号且作为oprater及参数的分隔符
            foo;;myvar ;;max([5,8,2,11, 7]);;rand(100)+120;;3+4;; 'singel;;quote' ;;"double;;quote" ;;`backquote;; ${bar}`

        2. 使用函数调用嵌套(todo:以后支持)
            foo(myvar,bar(zoo('fccdjny',6, 3.14),'a'), a+string(xxx))
    """
    # 不能简单使用字符串分割, 因为字符串中可能会有分隔符的模式出现, 比如:
    # foo ;; myvar ;; bar(zoo('fccd;;jny',6, 3.14),'a');; a+string(xxx)
    # foo;;myvar;;max([5, 8, 2, 11, 7]);;rand(100);; 'fccd;;jny' ;;3+4
    # expr.split(";;")
    ## todo: 是否将此模式编译
    pattern = r""";;(?=(?:[^"'`]*["'`][^"'`]*["'`])*[^"'`]*$)"""
    # To split the string by these semicolon:
    _parts = re.split(pattern, expr)
    parts = [i.strip() for i in _parts]  # 去掉表达式前后的空格
    return parts


import asyncio
import logging
from pathlib import Path
from pprint import pprint, pformat
import json
import contextvars
import pytest
import zen
logger = logging.getLogger(__name__)

# 设置 default 可以避免在调用 .get() 方法时报 LookupError 错误
# myvar = contextvars.ContextVar('myvar', default=None)
myvar = contextvars.ContextVar('myvar')


graph_json_customNode = """{
  "contentType": "application/vnd.gorules.decision",
  "nodes": [
    {
      "id": "115975ef-2f43-4e22-b553-0da6f4cc7f68",
      "type": "inputNode",
      "position": {
        "x": 180,
        "y": 240
      },
      "name": "Request"
    },
    {
      "id": "138b3b11-ff46-450f-9704-3f3c712067b2",
      "type": "customNode",
      "position": {
        "x": 470,
        "y": 240
      },
      "name": "customNode1",
      "content": {
        "kind": "sum",
        "config": {
          "prop1": "{{ a + 10 }}"
        }
      }
    },
    {
      "id": "db8797b1-bcc1-4fbf-a5d8-e7d43a181d5e",
      "type": "outputNode",
      "position": {
        "x": 780,
        "y": 240
      },
      "name": "Response"
    }
  ],
  "edges": [
    {
      "id": "05740fa7-3755-4756-b85e-bc1af2f6773b",
      "sourceId": "115975ef-2f43-4e22-b553-0da6f4cc7f68",
      "type": "edge",
      "targetId": "138b3b11-ff46-450f-9704-3f3c712067b2"
    },
    {
      "id": "5d89c1d6-e894-4e8a-bd13-22368c2a6bc7",
      "sourceId": "138b3b11-ff46-450f-9704-3f3c712067b2",
      "type": "edge",
      "targetId": "db8797b1-bcc1-4fbf-a5d8-e7d43a181d5e"
    }
  ]
}"""

graphs = {"custom_node_demo.json": graph_json_customNode}


def loader(key):
    return graphs[key]


async def custom_async_handler(request):
    """
        一旦有异步自定义函数，需要使用异步方法去求表达式的值
    """
    # v = myvar.get()
    try:
        v = myvar.get()
        logger.warning(f"myvar.get():{v}")
    except LookupError as e:
        logger.error(e, exc_info=True)
        raise e
    logger.warning(f"request.input:{pformat(request.input)}")
    logger.warning(f"request.node: {pformat(request.node)}")
    request_node = request.input.get("$nodes", {}).get("Request", {})
    logger.warning(f"get request node input:{pformat(request_node)}")
    prop1 = request.get_field("prop1")
    prop1_raw = request.get_field_raw("prop1")

    await asyncio.sleep(0)
    return {
        "output": {
                    "prop1_raw": prop1_raw,
                    "prop1": prop1,
                    "request.input": request.input,
                    "request.node": request.node,
                }
    }

@pytest.mark.asyncio
async def test_custmer_node_contextvars1():
    """
        测试自定义节点函数中的 contextvars 能否取得值.
        在实例化 ZenEngine 之前设置值.
    """
    o = object()
    myvar.set(o)
    engine = zen.ZenEngine({"loader": loader, "customHandler": custom_async_handler})
    logger.warning(f"myvar.set():{o}")
    res = await engine.async_evaluate("custom_node_demo.json", {"a": 10, "ip": "192.168.0.39"})
    assert res["result"]["prop1_raw"] == '{{ a + 10 }}'
    # 如果修改了规则图中的自定义节点的表达式, 需要重新修改这里的断言条件.
    assert res["result"]["prop1"]     == res["result"]["request.input"]['a'] + 10


@pytest.mark.asyncio
async def test_custmer_node_contextvars2():
    """
        测试自定义节点函数中的 contextvars 能否取得值.
        在实例化 ZenEngine 之后设置值.
        注意, 经过对 zen-engine 版本的二分查找定位, 发现 zen-engine==0.38.4 是这个用例失败的第一个版本.
        pytest test_zen_ch_contextvars.py  -sxlv --log-cli-level=0
    """
    engine = zen.ZenEngine({"loader": loader, "customHandler": custom_async_handler})
    o = object()
    myvar.set(o)
    logger.warning(f"myvar.set():{o}")
    res = await engine.async_evaluate("custom_node_demo.json", {"a": 10, "ip": "192.168.0.39"})
    assert res["result"]["prop1_raw"] == '{{ a + 10 }}'
    # 如果修改了规则图中的自定义节点的表达式, 需要重新修改这里的断言条件.
    assert res["result"]["prop1"]     == res["result"]["request.input"]['a'] + 1




import contextvars
import logging

###  contextvars 无法传递至 zen-engine 的 customHandler 中, 所以, 连接句柄等信息.
httpsession = contextvars.ContextVar('httpsession')


def get_state(sess_name):
    """
        sess_name: 
        1. http_session
        2. redis_connection
        3. postgresql connection
        4. xdb
    """
    pass
    c = {}
    # app.state[sess_name]

    return c[sess_name]


import pyparsing as pp

import pyparsing as pp
from pyparsing import Word, nums, alphanums, Literal, Suppress, Group, delimitedList, Forward, Optional
from pyparsing import infixNotation, OpAssoc, oneOf
# for recursive infix notations, or those with many precedence levels, it is best to enable packrat parsing
pp.ParserElement.enablePackrat()


def test_infix_fn_call():
    LPAREN, RPAREN = map(pp.Suppress, "()")

    arith_expr= pp.Forward()

    var_name = pp.pyparsing_common.identifier()
    integer = pp.pyparsing_common.integer()
    fn_call = pp.Group(var_name + LPAREN - pp.Group(pp.Optional(pp.delimitedList(arith_expr))) + RPAREN)
    arith_operand =  integer |fn_call| var_name

    rel_comparison_operator = pp.oneOf("< > <= >=")
    eq_comparison_operator = pp.oneOf("== !=")
    plus_minus_operator = pp.oneOf("+ -")
    mult_div_operator = pp.oneOf("* / %")
    ### **注意 表达式解析顺序.**
    ### 看样子是 arith_operand 中的 fn_call 和 var_name 的顺序会影响诸如 a + string(xxx) 的解析.
    ### 毕竟, fn_call 的定义格式中包含 var_name, 如果先解析 var_name, 将会导致无法识别 fn_call 格式.
    infix_expr = pp.infixNotation(arith_operand,
                                    [
                                        # add other operators here - in descending order of precedence
                                        # http://www.tutorialspoint.com/cprogramming/c_operators_precedence.htm
                                        (mult_div_operator, 2, pp.opAssoc.LEFT,),
                                        (plus_minus_operator, 2, pp.opAssoc.LEFT,),
                                        (rel_comparison_operator, 2, pp.opAssoc.LEFT,),
                                        (eq_comparison_operator, 2, pp.opAssoc.LEFT,),
                                    ]
                    )
    arith_expr <<= infix_expr


    test_string = "foo(bar()+fccdjny, 3)"
    parsed_result = arith_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "a + string(xxx)"
    parsed_result = arith_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "foo(33 + (a+ string(xxx)), 3)"
    parsed_result = arith_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    print(parsed_result.dump())
    tests = """\
        cos(60)
        sqrt(1 - sin(60) * sin(60))
        """
    arith_expr.runTests(tests)

    # divmod(a, 100)
    # iif(iif(condition1,value1,value2)>iif(condition2,value1,value2),value3,value4)



def test_infix_fn_call2():
    # Example base operands
    variable = pp.pyparsing_common.identifier
    integer = pp.pyparsing_common.integer

    # Example operator precedence list
    operator_precedence = [
        (Literal('-'), 1, OpAssoc.RIGHT), # Unary minus
        (oneOf('* /'), 2, OpAssoc.LEFT),
        (oneOf('+ -'), 2, OpAssoc.LEFT),
        (Literal('and'), 2, OpAssoc.LEFT),
        (Literal('or'), 2, OpAssoc.LEFT),
    ]

    # Combine base operand and operators to create the full expression grammar
    expression = Forward() # Use Forward for recursive definitions (e.g., nested expressions)

    # Example incorporating function calls
    LPAR, RPAR = map(Suppress, "()")
    func_name = Word(alphanums + '_')
    func_call = Group(func_name("name") + LPAR + Group(Optional(delimitedList(expression)))("args") + RPAR)

    # Update the base operand to include function calls
    expression <<= infixNotation(func_call | integer | variable, operator_precedence)

    # Example usage
    test_string = "f(x*y, g(z, y))"
    parsed_result = expression.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")
    # print(parsed_result.dump())

    test_string = "10 + var * 5 - 2"
    parsed_result = expression.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")
    # print(parsed_result.dump())

    test_string = "foo(bar, 3)"
    parsed_result = expression.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")
    # print(parsed_result.dump())

    test_string = "foo(zoo(bar), 3)"
    parsed_result = expression.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")
    # print(parsed_result.dump())

    test_string = "foo(bar()+fccdjny, 3)"
    parsed_result = expression.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")
    # print(parsed_result.dump())

    # Example usage
    test_string = "func1(x + y, func2(z)) * 5"
    parsed_result = expression.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    # Example usage
    test_string = "z * 5"
    parsed_result = expression.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")


if __name__ == "__main__":
    test_infix_fn_call()
    test_infix_fn_call2()


# Note: In pyparsing versions 3.0.x and later, operatorPrecedence has been renamed to infixNotation (and more 
# recently infix_notation following PEP 8 naming conventions). While operatorPrecedence might still work as 
# a compatibility synonym in some older 2.x versions, it is recommended to use infixNotation or infix_notation 
# for current and future code.

import pyparsing as pp
from pyparsing import oneOf, opAssoc, infixNotation, QuotedString

def test_infix_notation():
    # Define basic elements
    number = pp.pyparsing_common.number().setName("number")
    variable = pp.Word(pp.alphas, pp.alphanums + '_')
    double_quotes = QuotedString('"', escChar='\\', unquoteResults=False)
    single_quotes = QuotedString("'", escChar='\\', unquoteResults=False)
    string = (single_quotes | double_quotes).setName("string")
    # boolean = (pp.Literal("true") | pp.Literal("false")).setName("boolean")

    atom = number | variable | string 

    # Define operator precedence
    infix_expr = pp.infixNotation(atom, [
        ('-', 1, opAssoc.RIGHT),
        ('^', 2, pp.opAssoc.RIGHT),  # Exponentiation
        (pp.oneOf('* / %'), 2, pp.opAssoc.LEFT),  # Multiplication, Division, Modulo
        (pp.oneOf('+ -'), 2, pp.opAssoc.LEFT),  # Addition, Subtraction
        (pp.oneOf('> >= < <= == !='), 2, pp.opAssoc.LEFT),  # Comparison operators
        ('not', 1, pp.opAssoc.RIGHT),  # Logical NOT
        ('and', 2, pp.opAssoc.LEFT),  # Logical AND
        ('or', 2, pp.opAssoc.LEFT),  # Logical OR
    ])

    # Example usage
    # result = expression.parseString("2 + 3 * 4")
    # print(result) # Expected: ['2', '+', ['3', '*', '4']]

    # result = expression.parseString("True and False or not True")
    # print(result) # Expected: [['True', 'and', 'False'], 'or', ['not', 'True']]
    # # Example usage
    # result = infix_expr.parse_string("2 + 3 * 4")
    # print(result.as_list())  # Output: ['2', ['3', '*', '4'], '+'] (or similar structured output)

    test_string = "1"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "true"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "'true'"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")


    test_string = "1e+3"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "1e1 ^ 2"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "(1e2 + 2e2) * 1e1"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "-2 + 3 * 4 ^ 2 + (3 + 0.1) * 2"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "-5 * -5"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "10 % 3 != 0"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "5^2 + 3 * 4 - 6 / 2 > 20"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "x or not true"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "true"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "true and false"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "true and (not false or variable_name)"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")


    ## 字符串相加
    test_string = "'a' + 'b'"
    parsed_result = infix_expr.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")


if __name__ == "__main__":
    test_infix_notation()


import zen
import csv
import pytest
from pathlib import Path 

base_data_dir = Path(__file__).parent / "data"

def read_csv(file_path):
    data = []
    with open(file_path, 'r', encoding='utf-8') as file:
        reader = csv.reader(file, delimiter=';')
        next(reader)  # 跳过标题行
        for row in reader:
            # 跳过空行和带#的描述行
            if not row or row[0].strip().startswith('#'):
                continue
            if len(row) == 3:
                # 提取有效行数据（expression、input、output）
                expression = row[0].strip()
                input_data = row[1].strip()
                output_data = row[2].strip()
                data.append({
                    'expression': expression,
                    'input': input_data,
                    'output': output_data
                })
            elif len(row) == 2:
                # 提取有效行数据（expression、output）
                expression = row[0].strip()
                output_data = row[1].strip()
                data.append({
                    'expression': expression,
                    'output': output_data
                })
    return data


@pytest.fixture()
def parsed_date_csv():
    file_path  = base_data_dir / "date.csv"
    parsed_data = read_csv(file_path)
    return parsed_data


@pytest.fixture()
def parsed_standard_csv():
    file_path  = base_data_dir / "standard.csv"
    parsed_data = read_csv(file_path)
    return parsed_data


@pytest.fixture()
def parsed_unary_csv():
    file_path  = base_data_dir / "unary.csv"
    parsed_data = read_csv(file_path)
    return parsed_data


def test_zen_date_expression(parsed_date_csv):
    for data in parsed_date_csv:
        input = data.get("input", "")
        expression = data["expression"]
        output = data["output"] if data["output"] else None
        output = zen.evaluate_expression("d({})".format(output)) if "Z" in output else zen.evaluate_expression(output)
        input = zen.evaluate_expression(input) if input else None
        re = zen.evaluate_expression(expression, input)
        assert re == output


def test_zen_standard_expression(parsed_standard_csv):
    for data in parsed_standard_csv:
        input = data.get("input", "")
        expression = data["expression"]
        input = zen.evaluate_expression(input) if input else None
        output = data["output"] if data["output"] else None
        re = zen.evaluate_expression(expression, input)
        assert re == zen.evaluate_expression(output)


def test_zen_unary_expression(parsed_unary_csv):
    for data in parsed_unary_csv:
        input = data.get("input", "")
        expression = data["expression"]
        input = zen.evaluate_expression(input) if input else None
        output = zen.evaluate_expression(data["output"]) if data["output"] else None
        re = zen.evaluate_unary_expression(expression, input)
        assert re == output



import logging
for k,v in  logging.Logger.manager.loggerDict.items()  :
        print('+ [%s] {%s} ' % (str.ljust( k, 20)  , str(v.__class__)[8:-2]) )
        if not isinstance(v, logging.PlaceHolder):
            for h in v.handlers:
                print('     +++',str(h.__class__)[8:-2] )
import inspect
# print(inspect.getsource(logging))
logging.getLogger('foo').addHandler(logging.NullHandler())
logger=logging.getLogger('logger')
logger.debug('The system may break down debug')
logger.info('The system may break down info')
logger.warning('The system may break down warning')
logger.error('The system may break down error')
for k,v in  logging.Logger.manager.loggerDict.items()  :
        print('+ [%s] {%s} ' % (str.ljust( k, 20)  , str(v.__class__)[8:-2]) )
        if not isinstance(v, logging.PlaceHolder):
            for h in v.handlers:
                print('     +++',str(h.__class__)[8:-2] )


### 用下列命令来查看运行结果.
### pytest  tests/bench/test_zen_engine.py --benchmark-only
import random
import json
import string
from pprint import pprint
import pytest

import zen

con = """
{
  "contentType": "application/vnd.gorules.decision",
  "nodes": [
    {
      "type": "inputNode",
      "content": {
        "schema": ""
      },
      "id": "d4c5fdef-96c0-4c5e-8e7b-e1e4f9797251",
      "name": "request",
      "position": {
        "x": 245,
        "y": 355
      }
    },
    {
      "type": "outputNode",
      "content": {
        "schema": ""
      },
      "id": "a2cb6433-140b-4d2c-89f2-0e4f3140dd1a",
      "name": "response",
      "position": {
        "x": 1120,
        "y": 365
      }
    },
    {
      "type": "expressionNode",
      "content": {
        "expressions": [
          {
            "id": "8bf2c094-3115-4608-9416-1c854c68cc89",
            "key": "result",
            "value": "num * 2"
          }
        ],
        "passThrough": false,
        "inputField": null,
        "outputPath": null,
        "executionMode": "single"
      },
      "id": "45bbd693-493b-415d-8ad6-0d57c9a40d3e",
      "name": "expression1",
      "position": {
        "x": 715,
        "y": 400
      }
    }
  ],
  "edges": [
    {
      "id": "84fa9a9a-e90a-4322-be6f-94d3efd84edf",
      "sourceId": "d4c5fdef-96c0-4c5e-8e7b-e1e4f9797251",
      "type": "edge",
      "targetId": "45bbd693-493b-415d-8ad6-0d57c9a40d3e"
    },
    {
      "id": "0cd7ec7d-05d9-4497-abc7-d51b2202ddb4",
      "sourceId": "45bbd693-493b-415d-8ad6-0d57c9a40d3e",
      "type": "edge",
      "targetId": "a2cb6433-140b-4d2c-89f2-0e4f3140dd1a"
    }
  ]
}
"""

def init_zen():
  engine = zen.ZenEngine()
  decision = engine.create_decision(con)
  return engine, decision


engine, decision = init_zen()


def zenengine_demo0(con):
  engine = zen.ZenEngine()
  decision = engine.create_decision(con)
  input = {
    "num": 10,
  }
  result = decision.evaluate(input, {"trace": True})


def zenengine_demo1(con, engine):
  decision = engine.create_decision(con)
  input = {
    "num": 10,
  }
  result = decision.evaluate(input, {"trace": True})


def zenengine_demo2(con, decision):
  input = {
    "num": 10,
  }
  result = decision.evaluate(input, {"trace": True})


def zenengine_demo3(con, decision):
  input = {
    "num": 10,
  }
  result = decision.evaluate(input)


def zenengine_demo4(con, decision):
  input = {
    "num": 10,
    "extra_num": random.randint(0, 100)
  }
  result = decision.evaluate(input)


def zenengine_demo5(con, decision):
  input = {
    "num": 10,
  }
  key = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8))
  value = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8))
  input[key] = value
  result = decision.evaluate(input)


@pytest.mark.benchmark(
    group="zen-engine-decision",
    disable_gc=True,
    # warmup=False
)
def test_0(benchmark):
  # benchmark something
  result = benchmark(zenengine_demo0, con)

  # Extra code, to verify that the run completed correctly.
  # Sometimes you may want to check the result, fast functions
  # are no good if they return incorrect results :-)
  assert True


@pytest.mark.benchmark(
    group="zen-engine-decision",
    disable_gc=True,
    # warmup=False
)
def test_1(benchmark):
  # benchmark something
  result = benchmark(zenengine_demo1, con, engine)

  # Extra code, to verify that the run completed correctly.
  # Sometimes you may want to check the result, fast functions
  # are no good if they return incorrect results :-)
  assert True


@pytest.mark.benchmark(
    group="zen-engine-decision",
    disable_gc=True,
    # warmup=False
)
def test_2(benchmark):
  # benchmark something
  result = benchmark(zenengine_demo2, con, decision)

  # Extra code, to verify that the run completed correctly.
  # Sometimes you may want to check the result, fast functions
  # are no good if they return incorrect results :-)
  assert True


@pytest.mark.benchmark(
    group="zen-engine-decision",
    disable_gc=True,
    # warmup=False
)
def test_3(benchmark):
  # benchmark something
  result = benchmark(zenengine_demo3, con, decision)

  # Extra code, to verify that the run completed correctly.
  # Sometimes you may want to check the result, fast functions
  # are no good if they return incorrect results :-)
  assert True


@pytest.mark.benchmark(
    group="zen-engine-decision",
    disable_gc=True,
    # warmup=False
)
def test_4(benchmark):
  # benchmark something
  result = benchmark(zenengine_demo4, con, decision)

  # Extra code, to verify that the run completed correctly.
  # Sometimes you may want to check the result, fast functions
  # are no good if they return incorrect results :-)
  assert True


@pytest.mark.benchmark(
    group="zen-engine-decision",
    disable_gc=True,
    # warmup=False
)
def test_5(benchmark):
  # benchmark something
  result = benchmark(zenengine_demo4, con, decision)

  # Extra code, to verify that the run completed correctly.
  # Sometimes you may want to check the result, fast functions
  # are no good if they return incorrect results :-)
  assert True


def custom_handler(request):
    from pprint import pprint
    pprint(dir(request))
    print("request.input:  ", request.input)
    print("request.node:   ", request.node)

    func1 = lambda x: x
    func2 = lambda x: x
    customnode2functions = {
      "138b3b11-ff46-450f-9704-3f3c712067b2": func1,
      "8284e0b9-d3ba-4617-8a6c-86682ee24425": func2,

    }
    node_id = ""
    customnode2functions[node_id]
    # if request.node["id"] == node_id:
    #   func = customnode2functions[node_id]
    #   func(request.input)
    return {
        "output": { "sum": 100 }
    }


if __name__ == "__main__":
    engine = zen.ZenEngine()
    decision = engine.create_decision(con)
    input = {
      "num": 10,
    }
    res = decision.evaluate(input, {"trace": False})
    pprint(res)


import re

text = 'item1, "string with, a comma", item2, "another string", item3'

# Regex to find commas not inside double quotes
# This regex works by matching a comma that is NOT followed by an even number of quotes
# until the end of the string.
# It essentially checks if there's an odd number of quotes after the comma,
# implying it's inside a quoted string.
# This pattern is more robust for simple cases without escaped quotes inside strings.
pattern = r',(?=(?:[^"]*"[^"]*")*[^"]*$)'

# Find all matches
commas_outside_quotes = re.findall(pattern, text)

print(f"Commas found outside quotes: {commas_outside_quotes}")

# To split the string by these commas:
parts = re.split(pattern, text)
print(f"Split parts: {parts}")


# Explanation of the regex ,(?=(?:[^"]*"[^"]*")*[^"]*$):
# ,: Matches a literal comma.
# (?=...): This is a positive lookahead assertion. It means "assert that what follows matches the pattern inside, but don't consume characters."
# (?:[^"]*"[^"]*")*: This is a non-capturing group (?:...) that matches an even number of quoted sections.
# [^"]*: Matches any character except a double quote, zero or more times.
# ": Matches a double quote.
# This sequence [^"]*"[^"]*" effectively matches a complete quoted string (e.g., "abc" or "abc,def"). The * after the non-capturing group allows for zero or more such complete quoted sections.
# [^"]*$: Matches any character except a double quote, zero or more times, until the end of the string $.


from pathlib import Path
from pprint import pformat
import logging
import asyncio
import glob
import json
import os.path
import time
import unittest
import zen
logger = logging.getLogger(__name__)

basedir = Path(__file__).parent

async def loader(key):
    with open(basedir / "test-data/" / key, "r") as f:
        return f.read()


def sync_loader(key):
    with open(basedir / "test-data/" / key, "r") as f:
        return f.read()


def graph_loader(key):
    with open(basedir / "test-data/graphs/" / key, "r") as f:
        return f.read()


def custom_handler(request):
    p1 = request.get_field("prop1")
    return {
        "output": {"sum": p1}
    }


async def custom_async_handler(request):
    p1 = request.get_field("prop1")
    await asyncio.sleep(0.1)
    return {
        "output": {"sum": p1}
    }


class AsyncZenEngine(unittest.IsolatedAsyncioTestCase):
    def test_sync_loader_and_get_decision(self):
        engine = zen.ZenEngine({"loader": sync_loader})
        r1 = engine.get_decision("function.json")
        r2 = engine.get_decision("function.json")
        r3 = engine.get_decision("function.json")
        r4 = engine.get_decision("function.json")

    async def test_async_evaluate(self):
        engine = zen.ZenEngine({"loader": loader})
        r1 = engine.async_evaluate("function.json", {"input": 5})
        r2 = engine.async_evaluate("table.json", {"input": 2})
        r3 = engine.async_evaluate("table.json", {"input": 12})

        results = await asyncio.gather(r1, r2, r3)
        self.assertEqual(results[0]["result"]["output"], 10)
        self.assertEqual(results[1]["result"]["output"], 0)
        self.assertEqual(results[2]["result"]["output"], 10)

    async def test_async_evaluate_custom_handler(self):
        engine = zen.ZenEngine({"loader": loader, "customHandler": custom_async_handler})
        r1 = engine.async_evaluate("custom.json", {"a": 10})
        r2 = engine.async_evaluate("custom.json", {"a": 20})
        r3 = engine.async_evaluate("custom.json", {"a": 30})

        results = await asyncio.gather(r1, r2, r3)
        self.assertEqual(results[0]["result"]["sum"], 20)
        self.assertEqual(results[1]["result"]["sum"], 30)
        self.assertEqual(results[2]["result"]["sum"], 40)

    async def test_async_sleep_function(self):
        engine = zen.ZenEngine({"loader": loader, "customHandler": custom_async_handler})

        await engine.async_evaluate("sleep-function.json", {})
        self.assertTrue(True)

    async def test_async_http_function(self):
        engine = zen.ZenEngine({"loader": loader, "customHandler": custom_async_handler})

        await engine.async_evaluate("http-function.json", {})
        self.assertTrue(True)

    async def test_create_decisions_from_content(self):
        engine = zen.ZenEngine()
        with open(basedir / "test-data/function.json", "r") as f:
            functionContent = f.read()
        functionDecision = engine.create_decision(functionContent)

        r = await functionDecision.async_evaluate({"input": 15})
        self.assertEqual(r["result"]["output"], 30)

    async def test_evaluate_graphs(self):
        engine = zen.ZenEngine({"loader": graph_loader})
        json_files = glob.glob(str(basedir / "test-data/graphs/*.json"))
        # json_files = glob.glob(str(basedir / "test-data/*.json"))


        for json_file in json_files:
            logger.info("-----------------------")
            logger.info(json_file)
            with open(json_file, "r") as f:
                json_contents = json.loads(f.read())

            for test_case in json_contents["tests"]:
                logger.info("********start**********")
                logger.info("test_case:\n", pformat(test_case))
                key = os.path.basename(json_file)
                logger.info("before engine.async_evaluate")
                engine_response = await engine.async_evaluate(key, test_case["input"])
                logger.info("after engine.async_evaluate")
                logger.info("before engine.get_decision")
                decision = engine.get_decision(key)  # engine.get_decision 目前还不支持返回异步对象, 所以 graph_loader 只能用同步方法.
                logger.info("after engine.get_decision")
                decision_response = await decision.async_evaluate(test_case["input"])

                self.assertEqual(engine_response["result"], test_case["output"])
                self.assertEqual(decision_response["result"], test_case["output"])
                logger.info("********end**********")
            logger.info("-----------------------\n\n")


if __name__ == '__main__':
    unittest.main()



import zen

zen.evaluate_expression("{a:1}", {"a":3, "b":"fccdjny", "c": {'name': 'Alice', 'age': 30}})
zen.evaluate_expression("{'a':1}", {"a":3, "b":"fccdjny", "c": {'name': 'Alice', 'age': 30}})
zen.evaluate_expression('{"a":1}', {"a":3, "b":"fccdjny", "c": {'name': 'Alice', 'age': 30}})


## 注意, zen context 的键值对中的键是字符串，这个字符串可以用单引号，双引号，也可以不用引号(像一个变量一样)
zen.evaluate_expression('{ firstName: "John", lastName: last }',  {"firstName":"fff", "lastName": "lll", "last":"fccdjny"})
zen.evaluate_expression("{ 'firstName': 'John', lastName: last }",  {"firstName":"fff", "lastName": "lll", "last":"fccdjny"})
zen.evaluate_expression('{ "firstName": "John", lastName: last }',  {"firstName":"fff", "lastName": "lll", "last":"fccdjny"})
zen.evaluate_expression('values({ firstName: "John", lastName: last })',  {"firstName":"fff", "lastName": "lll", "last":"fccdjny"})
zen.evaluate_expression('keys({ firstName: "John", lastName: last })',  {"firstName":"fff", "lastName": "lll", "last":"fccdjny"})


## 测试 zen 表达式 变量数组 切片   支持
zen.evaluate_expression('a[1]', {"a": [1,2,3,4,5],"b":"fccdjny"})
zen.evaluate_expression('a[1:]', {"a": [1,2,3,4,5],"b":"fccdjny"})
zen.evaluate_expression('a[1:2]', {"a": [1,2,3,4,5],"b":"fccdjny"})

## 测试 zen 表达式 变量字符串 切片  支持
zen.evaluate_expression('b[1]', {"a": [1,2,3,4,5],"b":"fccdjny"})
zen.evaluate_expression('b[1:]', {"a": [1,2,3,4,5],"b":"fccdjny"})
zen.evaluate_expression('b[1:2]', {"a": [1,2,3,4,5],"b":"fccdjny"})

## 测试 zen 表达式 常量数组 切片    支持
zen.evaluate_expression('[1,2,3,4,5][1]', {"a": [1,2,3,4,5],"b":"fccdjny"})
zen.evaluate_expression('[1,2,3,4,5][1:]', {"a": [1,2,3,4,5],"b":"fccdjny"})
zen.evaluate_expression('[1,2,3,4,5][1:2]', {"a": [1,2,3,4,5],"b":"fccdjny"})

## 测试 zen 表达式 常量字符串 切片  不支持
zen.evaluate_expression('"fccdjny"[1]', {"a": [1,2,3,4,5],"b":"fccdjny"})
zen.evaluate_expression('"fccdjny"[1:]', {"a": [1,2,3,4,5],"b":"fccdjny"})
zen.evaluate_expression('"fccdjny"[1:2]', {"a": [1,2,3,4,5],"b":"fccdjny"})



zen.evaluate_expression('a[3: 3+3]', {"a": "fccdjny"})
zen.evaluate_expression('a[3: 3+1.3]', {"a": "fccdjny"})  


zen.evaluate_expression('`User ${user.name} is ${user.age} years old`', {"user":{"name":"John","age":30}})
zen.evaluate_expression("'User ${user.name} is ${user.age} years old'", {"user":{"name":"John","age":30}})



zen.evaluate_expression('user.contacts[i].phone.p[0]', {"i":1,"user":{"contacts":[{"phone":"123-456-7890"}, {"phone":{"p":["456-123-7890"]}}]}})
zen.evaluate_expression('user.contacts [i].phone.p[0]', {"i":1,"user":{"contacts":[{"phone":"123-456-7890"}, {"phone":{"p":["456-123-7890"]}}]}})



zen.evaluate_expression('a? 1 :2 + 3', {"a": True}) == 1
zen.evaluate_expression('(a? 1 : 2) + 3', {"a": True}) == 4



zen.evaluate_expression('a? b ?? "bbb" :2 + 3', {"a": True})



from typing import Any, Optional, TypedDict, Literal, Awaitable, Union
import logging
import json
from pathlib import Path
from pprint import pprint, pformat
import asyncio
import inspect

import zen
from zen import ZenDecision
from .custom.udf_manager import udf_manager, udf, FuncArg, FuncRet
from .custom import op_args_combination, parse_oprator_expr_v3
# from zen import EvaluateResponse  # cannot import
zen_exprs_eval = lambda x, input: zen.evaluate_expression(x, input)
logger = logging.getLogger(__name__)

class EvaluateResponse(TypedDict):
    performance: str
    result: dict
    trace: dict


async def custom_async_handler(request):
    """
        示例自定义节点处理函数.
        todo: 是否需要包装一层异常捕获的逻辑.
    """
    # p1 = request.get_field("prop1")  # 没有prop1属性会报错.
    # await asyncio.sleep(0.1)
    logger.debug(f"request:{request}")
    logger.debug(f"request attrs:{dir(request)}")
    result = zen.evaluate_expression('rand(100)', request.input)
    logger.debug(f"return value:{result}")
    return {
        "output": {"sum": 112}
    }

def loader(key):
    """
        loader 如果 loader 是异步函数, 那么同步的 get_decision 会有问题.
        除非我们自己实现 zenRule 的 get_decision, 而不是去调用 zenEngine的 get_decision
        加载规则还是让客户自己选择实现, 然后调用 create_decision_with_cache_key 缓存下来即可.
        暂时loader选择使用同步函数.
        此方法需要被覆写.
    """
    basedir = Path(__file__).parent
    with open(basedir / "custom" / key, "r", encoding="utf8") as f:
        logger.warning(f"graph json: %s", basedir / "custom" / key)
        return f.read()


class ZenRule:
    def __init__(self, options: Optional[dict] = None) -> None: 
        # {"customHandler": self.custom_handler_v3, "loader": self.loader}
        if options:
            ### loader 如果是异步函数, ZenEngine 的 get_decision 方法就会 hang 住. 这个需要在 rust 中去调整.
            loader = options.get("loader")
            if loader and inspect.iscoroutinefunction(loader):
                raise RuntimeError("loader is not allowed using async def, please use sync function definition")
            if not options.get("customHandler"):
                options["customHandler"] = self.custom_handler_v3
            self.options = options
        else: # custom_async_handler self.custom_handler_v3
            self.options = {"customHandler": self.custom_handler_v3}
        self.engine = zen.ZenEngine(self.options)

        self.decision_cache = {}  # key -> zen decision instance
        self.custom_context = {}  # 自定义节点中的需要传入的上线文参数, 比如 trace_id 或者规则的相关元信息.

    def create_decision(self, content) -> ZenDecision:
        """
            如果不想 decision 被缓存, 那么请使用 create_decision 方法来获取 decision.
        """
        content_ = self.graph_addons(content)
        # logger.debug(f"after graph_addons: {content_}")
        return self.engine.create_decision(content_)

    def create_decision_with_cache_key(self, key, content) -> ZenDecision:
        """
            创建规则和修改规则都是使用此方法.
        """
        if self.decision_cache.get(key):
            raise RuntimeError(f"rule key:{key} is existed, if confirm to overwrite this key,please use update_decision_with_cache_key")

        zendecision = self.create_decision(content)
        self.decision_cache[key] = zendecision
        return zendecision

    def update_decision_with_cache_key(self, key, content) -> ZenDecision:
        """
            创建规则和修改规则都是使用此方法.
        """
        if self.decision_cache.get(key):
            zendecision = self.create_decision(content)
            self.decision_cache[key] = zendecision
        else:
            raise RuntimeError(f"rule key:{key} is not existed, please use create_decision_with_cache_key")
        return zendecision

    def delete_decision_with_cache_key(self, key) -> None:
        """
            删除对应规则 decision 的缓存键.
        """
        if self.decision_cache.get(key):
            del self.decision_cache[key]
        else:
            raise RuntimeError(f"delete failed! rule key:{key} is not existed")

    def get_decision(self, key) -> ZenDecision:
        zendecision = self.decision_cache.get(key, None)
        if not zendecision:
            loader = self.options.get("loader")
            if not loader:
                raise RuntimeError(f"decision {key} not found, please create_decision_with_cache_key")
            decision_content = loader(key)
            zendecision = self.create_decision(decision_content)
            # zendecision = self.engine.get_decision(key)  # 会隐式调用 loader 函数.
            self.decision_cache[key] = zendecision
            return zendecision
        else:
            return zendecision

    def evaluate(self, key, ctx, options=None) -> EvaluateResponse:
        # return self.engine.evaluate(key, ctx, options)  # engine.evaluate 会隐式调用 loader 函数.
        decision = self.get_decision_cache(key)
        logger.debug(f"evaluate decision: {decision}")
        return decision.evaluate(ctx, options)

    def async_evaluate(self, key, ctx, options=None) -> Awaitable[EvaluateResponse]:
        # return self.engine.async_evaluate(key, ctx, options)  # engine.async_evaluate 会隐式调用 loader 函数.
        decision = self.get_decision_cache(key)
        # decision = self.engine.get_decision(key)
        logger.debug(f"async_evaluate decision: {decision}")
        return decision.async_evaluate(ctx, options)

    def _custom_node_v1_to_v3(self, node):
        v3_funcs = node["content"]["config"].get("expressions", [])
        v3_func_ids = {i["id"] for i in v3_funcs}
        v1_funcs = node["content"]["config"].get("inputs", [])
        v1_func_ids = {i["id"] for i in v1_funcs}

        if not v3_funcs:
            node["content"]["config"]["expressions"] = []
        if not v1_funcs:
            node["content"]["config"]["inputs"] = []

        for func_item in node["content"]["config"]["inputs"]:  # 遍历 v1 的输入函数 v3_func_ids
            func_name = func_item["funcmeta"]["name"]
            v1_arg_exprs = func_item["arg_exprs"]
            args = [[i["arg_name"], v1_arg_exprs[i["arg_name"]]] for i in func_item["funcmeta"]["arguments"]]
            args_ = ";;".join([j for _, j in args])
            func_call = f"{func_name};;{args_}"
            d = {
                "id": func_item["id"],
                "key": func_item["key"],
                "value": func_call
            }
            ## 如果当前 v1_func 的 id 不在 v3_func_ids 中, 那么就添加到 v3 的func表达式中
            if func_item["id"] not in v3_func_ids:
                node["content"]["config"]["expressions"].append(d)
            else:
                # 如果在 v3_func_ids 中, 那么更新此 v3_func
                for item in node["content"]["config"]["expressions"]:
                    if item["id"] == func_item["id"]:
                        item.update(d)

    def _custom_node_v3_to_v1(self, node):
        v3_funcs = node["content"]["config"].get("expressions", [])
        v3_func_ids = {i["id"] for i in v3_funcs}
        v1_funcs = node["content"]["config"].get("inputs", [])
        v1_func_ids = {i["id"] for i in v1_funcs}

        if not v3_funcs:
            node["content"]["config"]["expressions"] = []
        if not v1_funcs:
            node["content"]["config"]["inputs"] = []

        for func_item in node["content"]["config"]["expressions"]:
            expr_ast = parse_oprator_expr_v3(func_item["value"])
            operator, *op_arg_expressions  = expr_ast
            func_name = operator
            f = udf_manager.udf_info(func_name)  # 需要在这里设计一个函数执行异常时返回的值么.
            # args = [zen_exprs_eval(i, node_input) for i in op_arg_expressions]
            oprator_kwargs = op_args_combination(op_arg_expressions, f)
            d = {
                "id": func_item["id"],
                "key": func_item["key"],
                "type": "function",
                "arg_exprs": oprator_kwargs,
                "funcmeta": {
                    "name": f["name"],
                    "arglength": f["arglength"],
                    "arguments": f["arguments"],
                    "return_values": f["return_values"],
                    "comments": f["comments"]
                }
            }
            if func_item["id"] not in v1_func_ids:
                node["content"]["config"]["inputs"].append(d)
            else:
               for item in node["content"]["config"]["inputs"]:
                    if item["id"] == func_item["id"]:
                        item.update(d)

    def graph_addons(self, graph_content):
        if isinstance(graph_content, dict):
            rule_graph = graph_content
        elif isinstance(graph_content, str):
            try:
                rule_graph = json.loads(graph_content)
            except Exception as e:
                raise ValueError(f"Invalid JSON string: {e}")
        else:
            raise TypeError(f"Expected str or dict, got {type(graph_content).__name__}")
        ## 在 loader 和 create_decision 中隐式调用.
        ### 1.讲 inputNode 的 name 写到所有的customNode(自定义节点)中, 这样方便在自定义节点取得入参. 有些参数希望全局可以访问.
        input_node_name = [i.get("name") for i in rule_graph["nodes"] if i.get("type") == "inputNode"][0]
        for node in rule_graph["nodes"]:
            if node.get("type") == "customNode":
                content = node.get("content", {})
                config  = content.get("config", {})
                meta = config.get("meta", {})
                meta["inputNode_name"] = input_node_name
                node["content"]["config"]["meta"] = meta

                node["content"]["config"]["version"] = "v3"
                # # 新版编辑器完全上线后，需要线下把所有v1格式转化为 v3 格式.
                ## 只需要将 v1 格式增量转化为 v3 格式, 目前前端实现了v1和v3的兼容, 所以后端不需要将 v3 格式转化为 v1.
                self._custom_node_v1_to_v3(node)
                # # 旧版本编辑器移除后, 需要线下把所有v1格式转化为 v3 格式，并移除此代码(顺便可以去掉v1格式, 即删除content config inputs 字段).
                # self._custom_node_v3_to_v1(node)

                ### 将自定义节点中的表达式进行解析, 解析出其中表达式函数中的自定义函数(udf)的执行逻辑, 执行顺序.
                expr_asts = []
                custom_expressions = node["content"]["config"].get("expressions")
                if custom_expressions:
                    for func_item in custom_expressions:
                        item = {**func_item}
                        item["value"] = parse_oprator_expr_v3(func_item["value"])
                        expr_asts.append(item)
                    node["content"]["config"]["expr_asts"] = expr_asts

        logger.debug(f"rule_graph:{pformat(rule_graph)}")
        return json.dumps(rule_graph)

    @classmethod
    async def custom_handler_v1(cls, request):
        logger.debug("custom node use custom_handler_v1")
        funcs = request.node["config"].get("inputs", [])
        trans_func = lambda x: zen.evaluate_expression(x, request.input)
        res = {}
        bar = []
        for item in funcs:
            funcmeta = item["funcmeta"]
            funcname = funcmeta["name"]
            func_variables = {k: v for k, v in item["arg_exprs"].items()}
            vars2value = {k: trans_func(v) for k, v in func_variables.items()}
            env = {
                "node_id": request.node["id"],  ## 隔离 graph 中的节点
                "func_id": item["id"],  ## 隔离每一个计算逻辑.
                "meta": request.node["config"].get("meta", {}),
                **vars2value,
            }
            # env.update(vars2value)
            bar.append((item["key"], funcname, env))
        results = await asyncio.gather(*[udf_manager.call_udf(_[1], *_[2], **_[2]) for _ in bar])
        for key, result in zip([_[0] for _ in bar], results):
            res[key] = result
        res.update({k: v for k, v in request.input.items() if k != "$nodes"})
        logger.warning(f"custom v1 result:{res}")
        return {
            "output": res
        }


    @classmethod
    async def custom_handler_v3(cls, request):
        """
            v2 支持字面量的自定义函数的表达式, 和普通函数类似, 也支持函数表达式的嵌套调用.
            参考 src/custom/custom_v2.json 中的示例.
            兼容 custom_handler_v1 元数据定义和执行逻辑.

            为了简化决策引擎的使用, 决策引擎决定按照不同自定义节点算子的作用对自定义节点进行分类.
            v3 版本不支持 自定义算子 和 zen 表达式函数混合调用, 也不支持 自定义算子的嵌套调用.
            为了方便对自定义算子的参数中的复杂 zen 表达式进行解析，拟支持如下两种格式:

            foo;;myvar;;bar(zoo('fccdjny',6, 3.14),'a');; a+string(xxx)

            foo(myvar,bar(zoo('fccdjny',6, 3.14),'a'), a+string(xxx))

            这两种解析后得到的结果如下(json数组):
            ['foo', 'myvar', "bar(zoo('fccdjny',6, 3.14),'a')", ' a+string(xxx)']

            表示 foo 算子传入了 三个参数:
            1. zen 表达式变量 myvar
            2. zen 表达式函数 bar(zoo('fccdjny',6, 3.14),'a')
            3. zen 表达式 a+string(xxx)
            注意, 需要对 v1 版本和 v2 版本进行兼容.
            因为 v2 版本并未正式上线, 主要是对 v1 版本进行兼容.
        """
        # logger.debug(f"request.node:{request.node}")
        # graph json 要放在 zen engine zen rule 中进行解析, 解析的自定义表达式函数再使用自定义函数表达式来执行.
        expr_asts = request.node["config"].get("expr_asts", [])
        # if not expr_asts and request.node["config"].get("inputs"):  # 没有抽象语法树的解析, 那么使用 custom_handler_v1 版本.
        #     logger.debug("custom node use custom_handler_v1")
        #     return await cls.custom_handler_v1(request)
        logger.debug("custom node use custom_handler_v3")
        coro_funcs = []
        results = {}
        context = {
            "node_id": request.node["id"],  ## 隔离 graph 中的节点
            "meta": request.node["config"].get("meta", {}),
        }
        node_input_args = {k: v for k, v in request.input.items() if k != "$nodes"}
        for item in expr_asts:
            coro_funcs.append(cls.engine_v3(item, node_input_args, context))
        _results = await asyncio.gather(*coro_funcs)
        results = {k["key"]: v for k, v in zip(expr_asts, _results)}
        results.update(node_input_args)
        logger.debug(results)

        return {
            "output": results
        }

    @classmethod
    # async def engine_v3(cls, item, node_input, context):
    async def engine_v3(cls, exec_expr, node_input, context={}): 
        """
            item 代表自定义节点中的一个函数
            node_input 是此自定义节点的入参
            context 是额外传递的上下文(希望未来 zen-engine 直接支持此参数, 用于传入一些全局参数, 比如服务实例, trace_id等)

            exec_expr:
            ['foo', 'myvar', "bar(zoo('fccdjny',6, 3.14),'a')", ' a+string(xxx)']
        """
        try:
            expr_id = exec_expr["id"]
            expr_ast = exec_expr["value"]
            expr_key = exec_expr["key"]
            
            operator, *op_arg_expressions  = expr_ast
            logger.debug(f"node_input:{node_input}  context:{context}")
            logger.debug(f"operator:{operator}  args:{op_arg_expressions}")
            func_name = operator
            f = udf_manager.udf_info(func_name)  # 需要在这里设计一个函数执行异常时返回的值么.
            args = [zen_exprs_eval(i, node_input) for i in op_arg_expressions]
            oprator_kwargs = op_args_combination(args, f)
            logger.debug(f"ast_exec {func_name} args: {args}")
            logger.debug(f"ast_exec {func_name} kwargs: {oprator_kwargs}")
            kwargs = {
                **oprator_kwargs,
                **context,
                "func_id": expr_id,
                "expr_id": expr_id,
            }
            result = await udf_manager(func_name, *args, **kwargs)
            logger.debug(f"{func_name} calling ->: {result}")
            return result
        except Exception as e:
            logger.error(e, exc_info=True)
            # todo: 异常情况下自定义算子返回什么, 空字符串还是null.
            return None

    def get_decision_cache(self, key):
        return self.decision_cache.get(key, None)




__all__ = [
    ZenRule,
    udf,
    udf_manager,
    FuncArg,
    FuncRet,
]


# Ternary operator definition
# The '?' and ':' are defined as separate tokens, but they work together
# as part of a single ternary operation.
# The 3 indicates three operands.
# The OpAssoc.RIGHT is typical for ternary operators in many languages.

import pyparsing as pp

# Define basic elements
identifier = pp.pyparsing_common.identifier
integer = pp.pyparsing_common.integer
true_literal = pp.Literal("T")
false_literal = pp.Literal("F")
operand = integer | true_literal | false_literal | identifier

# Define the ternary operator using infixNotation
# The (("?", ":"), 3, pp.opAssoc.RIGHT) specifies the ternary operator
# with '?' and ':' as delimiters, 3 operands, and right-associativity.
ternary_expr = pp.infixNotation(
    operand,
    [   
        # Other operators can be defined here with their respective precedence
        # For example, binary arithmetic operators:
        # (pp.oneOf('* /'), 2, pp.OpAssoc.LEFT),
        # (pp.oneOf('+ -'), 2, pp.OpAssoc.LEFT),

        ((pp.Literal("?"), pp.Literal(":")), 3, pp.opAssoc.RIGHT),
    ]
)

# Example usage
test_string = "T ? 1 : F ? 2 : 3"
parsed_result = ternary_expr.parseString(test_string, parseAll=True)
# print(parsed_result.asList())
print(f"{test_string} --> {parsed_result}")


test_string = "T ? a : F ? 2 : 3"
parsed_result = ternary_expr.parseString(test_string, parseAll=True)
# print(parsed_result.asList())
print(f"{test_string} --> {parsed_result}")

test_string = "1 ? (2 ? 3 : 4) : 5"
parsed_result = ternary_expr.parseString(test_string)
print(f"{test_string} --> {parsed_result}")


import pyparsing as pp
from pyparsing import Literal, Word, delimited_list, Group, Forward, nums, QuotedString, Optional, alphanums, alphas, identchars, Suppress

# Define basic elements
# integer = Word(nums).set_parse_action(lambda t: int(t[0]))
integer = pp.pyparsing_common.number 
string = QuotedString('"') | QuotedString("'")
identifier = Word(alphas + '_' , alphanums + '_')

strslice = Forward()
_strslice = Group(
    identifier + Literal("[") + Optional(strslice) + Literal(":") + Optional(strslice) + Literal("]")
)
strslice <<= _strslice | integer | identifier

# # Example usage
# list_string = '[1, "hello", [2, "world"], 3]'
# parsed_list = py_list.parseString(list_string)
# print(parsed_list.asList())

# Example usage
test_string = "strings[1:2]"
parsed_result = strslice.parseString(test_string)
print(f"{test_string} --> {parsed_result}")

test_string = "strings[a:b]"
parsed_result = strslice.parseString(test_string)
print(f"{test_string} --> {parsed_result}")

test_string = "strings[a:2]"
parsed_result = strslice.parseString(test_string)
print(f"{test_string} --> {parsed_result}")


test_string = "strings[a:abc[2:2]]"
parsed_result = strslice.parseString(test_string)
print(f"{test_string} --> {parsed_result}")


import datetime

import pytest
import zen


def test_zen_template_str():
    inputs = {"a": 10, "b": "cjn","__meta__":{"redispool": "redispool", "httpclient": "httpclient"}}
    template = '{{ a + 10 }}'
    # 如果修改了规则图中的自定义节点的表达式, 需要重新修改这里的断言条件.

    # print("render_template:", zen.render_template("{{ a + 10 }}", inputs)) 
    # print("evaluate_expression:", zen.evaluate_expression("'{{ a + 10 }}'", inputs))  # res["result"]["prop1_raw"]
    aa = zen.render_template("{{ a + 10 }}", inputs)

    assert zen.render_template(template, inputs) == 20, "zen template addition operation return int type"
    assert zen.render_template("{}".format(template), inputs) == 20, "zen template addition operation return int type"
    assert zen.render_template("   {}   ".format(template), inputs) == 20, "zen 模板如果返回一个数字，即使前后都是空格，最后也会被隐式求值为数字"
    assert zen.render_template("a {}".format(template), inputs) == 'a 20', "zen template addition operation return int type"

    assert zen.evaluate_expression("'{}'".format(template), inputs) == '{{ a + 10 }}', "zen template 如果放在表达式中用引号括起来，会阻止模板渲染，仅作为字面字符串表示"  
    with pytest.raises(RuntimeError) as excinfo:
        assert zen.evaluate_expression("brde {{ a + 10 }}", inputs) == 20, "zen 模板不能直接在表达式中使用." 

    assert "parserError" in str(excinfo.value)


# zen expression 表达式语言测试

def test_zen_expression_str():
    '''
    字符串是一种表示文本内容（字符列表、单词）的数据类型。它通过使用单引号'或双引号括住字符来定义"。
    示例：
    "double quote string"
    'single quote string'
    '''
    # 字符串操作符号 +
    assert zen.evaluate_expression("a+b", { "a": "hello ", "b": "world" }) == "hello world"

    # 字符串slice [start, end]满足左闭右闭区间, 与python字符串切片不一样，python为左闭右开区间
    assert zen.evaluate_expression("a[0:5]", { "a": "0123456789abcdef"}) == '012345'

    # 变量等值与不等值判断 == !=
    assert zen.evaluate_expression("a == b", { "a": "hello ", "b": "world" }) == False
    assert zen.evaluate_expression("a != b", { "a": "hello ", "b": "world" }) == True

    # 字符串函数: len 接受一个字符串并返回其中的字符数
    assert zen.evaluate_expression("len(a) == 6", {"a": "string"}) == True

    # 字符串函数: upper 接受一个字符串并返回它的大写版本
    assert zen.evaluate_expression("upper(a)", {"a": "string"}) == 'STRING'

    # 字符串函数: lower 接受一个字符串并返回其小写版本
    assert zen.evaluate_expression("lower(a)", {"a": "StrIng"}) == 'string'

    # 字符串函数: startsWith 接受两个参数，如果字符串以指定值开头，则返回 true
    assert zen.evaluate_expression("startsWith(a, 'Sat')", {"a": "Saturday night plans"}) == True
    assert zen.evaluate_expression("startsWith(a, 'Sun')", {"a": "Saturday night plans"}) == False

    # 字符串函数: endsWith 接受两个参数，如果字符串以指定值结尾，则返回 true
    assert zen.evaluate_expression("endsWith(a, 'plans')", {"a": "Saturday night plans"}) == True
    assert zen.evaluate_expression("endsWith(a, 'night')", {"a": "Saturday night plans"}) == False

    # 字符串函数: contains 接受两个参数，如果字符串包含指定值，则返回 true
    assert zen.evaluate_expression("contains(a, 'night')", {"a": "Saturday night plans"}) == True
    assert zen.evaluate_expression("contains(a, 'plans')", {"a": "Saturday night plans"}) == True
    assert zen.evaluate_expression("contains(a, 'saturday')", {"a": "Saturday night plans"}) == False

    # 字符串函数: matches 接受两个参数，如果字符串与正则表达式匹配，则返回 true
    assert zen.evaluate_expression(r"matches(a, '^\d+$')", {"a": "12345"}) == True
    assert zen.evaluate_expression(r"matches(a, '^\d+$')", {"a": "12345a"}) == False

    # 字符串函数: extract 接受两个参数，返回正则表达式中捕获组的数组
    assert zen.evaluate_expression(r"extract(a, '(\d{4})-(\d{2})-(\d{2})')", {"a": "2022-02-01"}) ==  ['2022-02-01', '2022', '02', '01']
    assert zen.evaluate_expression(r"extract(a, '(\w+).(\w+)')", {"a": "foo.bar"}) == ['foo.bar', 'foo', 'bar']

    # 字符串函数: string 接受一个参数，尝试将变量转换为字符串，失败时抛出错误
    assert zen.evaluate_expression("string(a) == 'abc'", {"a": "abc"}) == True
    assert zen.evaluate_expression("string(a) == '20'", {"a": "20"}) == True
    assert zen.evaluate_expression("string(a) == '20'", {"a": 20}) == True
    assert zen.evaluate_expression("string(a) == 'true'", {"a": True}) == True  # ZEN Bool: true false


def test_zen_expression_num():
    '''
    数字可以是整数或小数(浮点数)。可以使用数字0-9, .（小数分隔符）和来定义, 使用_为了便于阅读
    示例：
    100;
    1_000_000;
    1.54;
    '''
    # 数字一元操作符号 -  取负数
    assert zen.evaluate_expression("-a", { "a": 3 }) == -3

    # 数字算术运算符： + - * / ^  %
    assert zen.evaluate_expression("a + b", {"a": 3, "b": -9}) == -6
    assert zen.evaluate_expression("a - b", {"a": 3, "b": 9}) == -6
    assert zen.evaluate_expression("a * b", {"a": 3, "b": 9}) == 27
    assert zen.evaluate_expression("b / a", {"a": 3, "b": 9}) == 3
    assert zen.evaluate_expression("a ^ b", {"a": 3, "b": 3}) == 27
    assert zen.evaluate_expression("a % b", {"a": 10, "b": 3}) == 1

    # 数字比较运算符： == != > < >= <=
    assert zen.evaluate_expression("a == 10", {"a": 10}) == True
    assert zen.evaluate_expression("a != 100", {"a": 10}) == True
    assert zen.evaluate_expression("a < 11", {"a": 10}) == True
    assert zen.evaluate_expression("a > 1", {"a": 10}) == True
    assert zen.evaluate_expression("a >= 10", {"a": 10}) == True
    assert zen.evaluate_expression("a <= 10", {"a": 10}) == True

    # 数字 函数
    # abs 接受一个数字，并返回其绝对值
    assert zen.evaluate_expression("abs(a) == 1.2", {"a": -1.2}) == True
    assert zen.evaluate_expression("abs(a) == 10", {"a": 10}) == True

    # rand 接受一个正数（限制），并返回 0 和提供的限制之间的生成数字
    assert zen.evaluate_expression("rand(a) < 4 ", {"a": 3}) == True
    assert zen.evaluate_expression("rand(a) in [0..3] ", {"a": 3}) == True

    # round 接受一个数字并将该数字四舍六入五取偶为最接近的整数
    assert zen.evaluate_expression("round(a) == 4", {"a":  3.5}) == True
    assert zen.evaluate_expression("round(a) == 3", {"a": 2.5}) == True
    assert zen.evaluate_expression("round(a) == 4", {"a": 4.49}) == True
    assert zen.evaluate_expression("round(a) == 5", {"a": 4.51}) == True

    # ceil 接受一个数字并将其向上舍入。它返回大于或等于给定数字的最小整数
    assert zen.evaluate_expression("ceil(a) == 5", {"a": 4.01}) == True
    assert zen.evaluate_expression("ceil(a) == 5", {"a": 4.9}) == True
    assert zen.evaluate_expression("ceil(a) == 4", {"a": 4.0}) == True

    # floor 接受一个数字并将其向下舍入。它返回小于或等于给定数字的最大整数
    assert zen.evaluate_expression("floor(a) == 4", {"a": 4.01}) == True
    assert zen.evaluate_expression("floor(a) == 4", {"a": 4.9}) == True
    assert zen.evaluate_expression("floor(a) == 4", {"a": 4.0}) == True

    # number 接受一个参数，尝试将数字或字符串转换为数字，失败时抛出错误
    assert zen.evaluate_expression("number(a) == 4", {"a": '4'}) == True
    assert zen.evaluate_expression("number(a) == 4", {"a": 4}) == True
    assert zen.evaluate_expression("number(a) == 1", {"a": True}) == True
    assert zen.evaluate_expression("number(a) == 0", {"a": False}) == True

    # isNumeric 接受一个参数，返回 bool，如果变量是数字或可以转换为数字的字符串则返回 true
    assert zen.evaluate_expression("isNumeric(a)", {"a": '4'}) == True
    assert zen.evaluate_expression("isNumeric(a)", {"a": 'test'}) == False
    assert zen.evaluate_expression("isNumeric(a)", {"a": True}) == False


def test_zen_expression_bool():
    '''
    布尔是一种逻辑数据类型，可以是true也可以是false。
    '''
    # 布尔逻辑运算符： and or ！not
    assert zen.evaluate_expression("a and b", { "a": True, "b": True}) == True
    assert zen.evaluate_expression("a and b", { "a": True, "b": False}) == False
    assert zen.evaluate_expression("a or b", { "a": True, "b": False}) == True
    assert zen.evaluate_expression("!a", { "a": False}) == True
    assert zen.evaluate_expression("not a", { "a": False}) == True
    
    # 布尔操作运算符： == !=
    assert zen.evaluate_expression("a == '10'", {"a": '10'}) == True
    assert zen.evaluate_expression("a != 11", {"a": 10}) == True

    # 三元操作符
    assert zen.evaluate_expression("product.price > 100 ? 'premium' : 'value'", {"product": {"price": 105}}) == 'premium'

    # 函数 bool 接受一个参数，尝试将变量转换为布尔值，失败时抛出错误
    assert zen.evaluate_expression("bool(a)", {"a": 0}) == False
    assert zen.evaluate_expression("bool(a)", {"a": 1}) == True
    assert zen.evaluate_expression("bool(a)", {"a": True}) == True
    assert zen.evaluate_expression("bool(a)", {"a": False}) == False
    assert zen.evaluate_expression("bool(a)", {"a": 'true'}) == True
    assert zen.evaluate_expression("bool(a)", {"a": 'false'}) == False
    assert zen.evaluate_expression("bool(a)", {"a": 'abc'}) == False

    
def test_zen_expression_date_and_time():
    '''
    日期、时间和持续时间是一组虚拟数据类型，内部使用 unix 时间戳（数字）表示为数字。
    '''
    # 日期时间函数： date time duration dayOfWeek dayOfMonth weekOfMonth weekOfYear seasonOfYear
    # date 接受格式化的字符串作为输入并以秒为单位返回 unix 时间戳
    assert zen.evaluate_expression("date('2022-01-01')") == 1640995200  # 2022-01-01对应的时间戳
    assert zen.evaluate_expression("date('2022-04-04T21:48:30Z')") == 1649108910  # 2022-04-04T21:48:30Z对应的时间戳
    assert zen.evaluate_expression("date('now')") == datetime.datetime.now().timestamp().__floor__()    # 当前时间戳

    # time 接受格式化的字符串作为输入并返回表示午夜(midnight)秒数的数字
    assert zen.evaluate_expression("time('21:49')") == 78540                    # 78540  从0时开始到21:49之间的秒数
    assert zen.evaluate_expression("time('21:48:20')") == 78500                 # 78500  从0时开始到21:48:20之间的秒数
    assert zen.evaluate_expression("time('2022-04-04T21:48:30Z')") == 78510     # 78510  从0时开始到21:48:30之间的秒数

    # duration 接受格式化的字符串（从秒到小时）作为输入并以秒为单位返回持续时间
    assert zen.evaluate_expression("duration('1h')") == 3600    # 3600    1小时包含的秒数
    assert zen.evaluate_expression("duration('30m')") == 1800   # 1800   30分钟包含的秒数
    assert zen.evaluate_expression("duration('10h')") == 36000  # 36000  10小时包含的秒数

    # dayOfWeek 接受时间戳并以数字形式返回星期几
    assert zen.evaluate_expression("dayOfWeek(date('2022-11-08'))") == 2  # 2   返回值为int

    # dayOfMonth 接受时间戳并以数字形式返回月份中的日期
    assert zen.evaluate_expression("dayOfMonth(date('2022-11-09'))") == 9  # 9   返回值为int

    # dayOfYear 接受时间戳并以数字形式返回一年中的某一天
    assert zen.evaluate_expression("dayOfYear(date('2022-11-10'))") == 314  # 314   返回值为int

    # weekOfMonth 接受时间戳并以数字形式返回月份的周数
    # zen-engine==0.24.2不支持内置的weekOfMonth方法
    # assert zen.evaluate_expression("weekOfMonth(date('2022-11-11'))") == 2  # 2   返回值为int

    # weekOfYear 接受时间戳并以数字形式返回一年中的第几周
    assert zen.evaluate_expression("weekOfYear(date('2022-11-12'))") == 45  # 2   返回值为int

    # seasonOfYear 接受 unix 时间戳并以字符串形式返回一年中的季节
    # zen-engine==0.24.2不支持内置的seasonOfYear方法
    # assert zen.evaluate_expression("seasonOfYear(date('2022-11-13'))") == 'Autumn'  # Autumn  返回值为str

    # monthString 接受时间戳并以字符串形式返回月份
    assert zen.evaluate_expression("monthString(date('2022-11-14'))") == 'Nov'  # Nov  返回值为str

    # weekdayString 接受时间戳并以字符串形式返回星期几
    assert zen.evaluate_expression("weekdayString(date('2022-11-14'))") == 'Mon'  # Mon  返回值为str

    # startOf 接受时间戳和单位。根据指定单位返回日期的开始时间
    assert zen.evaluate_expression(
        "startOf(date('2022-11-14 15:45:12'), 'second')") == 1668440712     # 1668440712  2022-11-14 15:45:12对应的时间戳 传递参数: "s" | "second" | "seconds"
    assert zen.evaluate_expression(
        "startOf(date('2022-11-14 15:45:12'), 'minute')") == 1668440700     # 1668384000  2022-11-14 15:45:00对应的时间戳 传递参数: "m" | "minute" | "minutes"
    assert zen.evaluate_expression(
        "startOf(date('2022-11-14 15:45:12'), 'hour')") == 1668438000       # 1668438000  2022-11-14 15:00:00对应的时间戳 传递参数: "h" | "hour" | "hours"
    assert zen.evaluate_expression(
        "startOf(date('2022-11-14 15:45:12'), 'day')") == 1668384000        # 1668384000  2022-11-14 00:00:00对应的时间戳 传递参数: "d" | "day" | "days"
    assert zen.evaluate_expression(
        "startOf(date('2022-11-14 15:45:12'), 'week')") == 1668384000       # 1668384000  2022-11-14 00:00:00对应的时间戳 传递参数: "w" | "week" | "weeks"
    assert zen.evaluate_expression(
        "startOf(date('2022-11-14 15:45:12'), 'month')") == 1667260800      # 1667260800  2022-11-01 00:00:00对应的时间戳 传递参数: "M" | "month" | "months"
    assert zen.evaluate_expression(
        "startOf(date('2022-11-14 15:45:12'), 'year')") == 1640995200       # 1640995200  2022-01-01 00:00:00对应的时间戳 传递参数: "y" | "year" | "years"

    # endOf 接受时间戳和单位。根据指定单位返回日期的结束时间
    assert zen.evaluate_expression(
        "endOf(date('2022-11-14 15:45:12'), 'second')") == 1668440712       # 1668440712  2022-11-14 15:45:12对应的时间戳 传递参数: "s" | "second" | "seconds"
    assert zen.evaluate_expression(
        "endOf(date('2022-11-14 15:45:12'), 'minute')") == 1668440759       # 1668440759  2022-11-14 15:45:59对应的时间戳 传递参数: "m" | "minute" | "minutes"
    assert zen.evaluate_expression(
        "endOf(date('2022-11-14 15:45:12'), 'hour')") == 1668441599         # 1668441599  2022-11-14 15:59:59对应的时间戳 传递参数: "h" | "hour" | "hours"
    assert zen.evaluate_expression(
        "endOf(date('2022-11-14 15:45:12'), 'day')") == 1668470399          # 1668470399  2022-11-14 23:59:59对应的时间戳 传递参数: "d" | "day" | "days"
    assert zen.evaluate_expression(
        "endOf(date('2022-11-14 15:45:12'), 'week')") == 1668988799         # 1668988799  2022-11-20 23:59:59对应的时间戳 传递参数: "w" | "week" | "weeks"
    assert zen.evaluate_expression(
        "endOf(date('2022-11-14 15:45:12'), 'month')") == 1669852799        # 1669852799  2022-11-30 23:59:59对应的时间戳 传递参数: "M" | "month" | "months"
    assert zen.evaluate_expression(
        "endOf(date('2022-11-14 15:45:12'), 'year')") == 1672444799         # 1672444799  2022-12-30 23:59:59对应的时间戳 传递参数: "y" | "year" | "years"

    # 将时间戳转化为日期字符串
    assert zen.evaluate_expression("dateString(date('2023-10-15'))") == '2023-10-15 00:00:00'
    assert zen.evaluate_expression("dateString(date('2023-10-15') + duration('12h'))") == '2023-10-15 12:00:00'
    assert zen.evaluate_expression("dateString(startOf('2023-01-01 15:45:01', 'day'))") == '2023-01-01 00:00:00'
    assert zen.evaluate_expression("dateString(endOf('2023-01-01 15:45:01', 'd'))") == '2023-01-01 23:59:59'
    assert zen.evaluate_expression("dateString(startOf('2023-01-01 15:45:01', 'hour'))") == '2023-01-01 15:00:00'
    assert zen.evaluate_expression("dateString(endOf('2023-01-01 15:45:01', 'h'))") == '2023-01-01 15:59:59'
    assert zen.evaluate_expression("dateString(startOf('2023-01-01 15:45:01', 'minute'))") == '2023-01-01 15:45:00'
    assert zen.evaluate_expression("dateString(endOf('2023-01-01 15:45:01', 'm'))") == '2023-01-01 15:45:59'
    assert zen.evaluate_expression("dateString(startOf('2023-01-04 15:45:01', 'week'))") == '2023-01-02 00:00:00'
    assert zen.evaluate_expression("dateString(endOf('2023-01-04 15:45:01', 'w'))") == '2023-01-08 23:59:59'
    assert zen.evaluate_expression("dateString(startOf('2023-01-04 15:45:01', 'month'))") == '2023-01-01 00:00:00'
    assert zen.evaluate_expression("dateString(endOf('2023-01-04 15:45:01', 'M'))") == '2023-01-31 23:59:59'
    assert zen.evaluate_expression("dateString(startOf('2023-01-04 15:45:01', 'year'))") == '2023-01-01 00:00:00'
    assert zen.evaluate_expression("dateString(endOf('2023-01-04 15:45:01', 'y'))") == '2023-12-31 23:59:59'


    # Date function basics zen-enign新增的时间函数
    assert zen.evaluate_expression("d('2023-10-15 14:30')") == '2023-10-15T14:30:00+08:00'    # 当前时区为
    assert zen.evaluate_expression("d('2023-10-15')") == '2023-10-15T00:00:00+08:00'
    assert zen.evaluate_expression("d('2023-10-15 14:30:45')") == '2023-10-15T14:30:45+08:00'
    assert zen.evaluate_expression("d('2023-10-15', 'Europe/Berlin')") == '2023-10-15T00:00:00+02:00'
    assert zen.evaluate_expression("d('2023-10-15 14:30', 'Europe/Berlin')") == '2023-10-15T14:30:00+02:00'
    assert zen.evaluate_expression("d('2023-10-15 14:30:45', 'Europe/Berlin')") == '2023-10-15T14:30:45+02:00'
    assert zen.evaluate_expression("d('Europe/Berlin').isValid() and d('Europe/Berlin').isToday()") == True

    # Date manipulation
    assert zen.evaluate_expression("d('2023-10-15').add('1d')") == '2023-10-16T00:00:00+08:00'
    assert zen.evaluate_expression("d('2023-10-15').add('1d 5h')") == '2023-10-16T05:00:00+08:00'
    assert zen.evaluate_expression("d('2023-10-15').add(1, 'd')") == '2023-10-16T00:00:00+08:00'
    assert zen.evaluate_expression("d('2023-10-15').sub('2d')") == '2023-10-13T00:00:00+08:00'
    assert zen.evaluate_expression("d('2023-10-15').sub(1, 'M')") == '2023-09-15T00:00:00+08:00'

    # Date comparisons
    assert zen.evaluate_expression("d('2023-10-15').isBefore(d('2023-10-16'))") == True
    assert zen.evaluate_expression("d('2023-10-15').isBefore(d('2023-10-16'))") == True
    assert zen.evaluate_expression("d('2023-10-15').isBefore(d('2023-10-15'))") == False
    assert zen.evaluate_expression("d('2023-10-15').isAfter(d('2023-10-14'))") == True
    assert zen.evaluate_expression("d('2023-10-15').isAfter(d('2023-10-15'))") == False
    assert zen.evaluate_expression("d('2023-10-15').isSame(d('2023-10-15'))") == True
    assert zen.evaluate_expression("d('2023-10-15').isSame(d('2023-10-16'))") == False
    assert zen.evaluate_expression("d('2023-10-15').isSameOrBefore(d('2023-10-15'))") == True
    assert zen.evaluate_expression("d('2023-10-15').isSameOrBefore(d('2023-10-16'))") == True
    assert zen.evaluate_expression("d('2023-10-15').isSameOrBefore(d('2023-10-14'))") == False
    assert zen.evaluate_expression("d('2023-10-15').isSameOrAfter(d('2023-10-15'))") == True
    assert zen.evaluate_expression("d('2023-10-15').isSameOrAfter(d('2023-10-14'))") == True
    assert zen.evaluate_expression("d('2023-10-15').isSameOrAfter(d('2023-10-16'))") == False
    assert zen.evaluate_expression("d('2025-02-01').isAfter(d('2025-01-05'), 'year')") == False
    assert zen.evaluate_expression("d('2025-02-01').isAfter(d('2024-12-31'), 'year')") == True
    assert zen.evaluate_expression("d('2025-02-01').isBefore(d('2025-03-01'), 'year')") == False
    assert zen.evaluate_expression("d('2025-02-01').isBefore(d('2026-01-01'), 'year')") == True
    assert zen.evaluate_expression("d('2025-02-01').isSame(d('2025-12-31'), 'year')") == True
    assert zen.evaluate_expression("d('2025-02-01').isSame(d('2026-01-01'), 'year')") == False
    assert zen.evaluate_expression("d('2025-02-15').isAfter(d('2025-01-05'), 'month')") == True
    assert zen.evaluate_expression("d('2025-02-15').isAfter(d('2025-02-05'), 'month')") == False
    assert zen.evaluate_expression("d('2025-02-15').isBefore(d('2025-03-01'), 'month')") == True
    assert zen.evaluate_expression("d('2025-02-15').isBefore(d('2025-02-28'), 'month')") == False
    assert zen.evaluate_expression("d('2025-02-15').isSame(d('2025-02-28'), 'month')") == True
    assert zen.evaluate_expression("d('2025-02-15').isSame(d('2025-03-01'), 'month')") == False
    assert zen.evaluate_expression("d('2025-04-15').isAfter(d('2025-01-01'), 'quarter')") == True
    assert zen.evaluate_expression("d('2025-02-15').isAfter(d('2025-03-31'), 'quarter')") == False
    assert zen.evaluate_expression("d('2025-02-15').isBefore(d('2025-04-01'), 'quarter')") == True
    assert zen.evaluate_expression("d('2025-02-15').isBefore(d('2025-03-31'), 'quarter')") == False
    assert zen.evaluate_expression("d('2025-02-15').isSame(d('2025-03-15'), 'quarter')") == True
    assert zen.evaluate_expression("d('2025-02-15').isSame(d('2025-04-01'), 'quarter')") == False
    assert zen.evaluate_expression("d('2025-02-15').isAfter(d('2025-02-14'), 'day')") == True
    assert zen.evaluate_expression("d('2025-02-15').isAfter(d('2025-02-15'), 'day')") == False
    assert zen.evaluate_expression("d('2025-02-15').isBefore(d('2025-02-16'), 'day')") == True
    assert zen.evaluate_expression("d('2025-02-15').isBefore(d('2025-02-15'), 'day')") == False
    assert zen.evaluate_expression("d('2025-02-15').isSame(d('2025-02-15 12:00:00'), 'day')") == True
    assert zen.evaluate_expression("d('2025-02-15').isSame(d('2025-02-16'), 'day')") == False

    # Date getters
    assert zen.evaluate_expression("d('2023-10-15').year()") == 2023
    assert zen.evaluate_expression("d('2023-10-15').month()") == 10
    assert zen.evaluate_expression("d('2023-10-15').day()") == 15
    assert zen.evaluate_expression("d('2023-10-15').weekday()") == 7
    assert zen.evaluate_expression("d('2023-10-16').weekday()") == 1
    assert zen.evaluate_expression("d('2023-10-15').hour()") == 0
    assert zen.evaluate_expression("d('2023-10-15').minute()") == 0
    assert zen.evaluate_expression("d('2023-10-15').second()") == 0
    assert zen.evaluate_expression("d('2023-10-15').dayOfYear()") == 288
    assert zen.evaluate_expression("d('2023-01-01').dayOfYear()") == 1
    assert zen.evaluate_expression("d('2023-12-31').dayOfYear()") == 365
    assert zen.evaluate_expression("d('2024-12-31').dayOfYear()") == 366
    assert zen.evaluate_expression("d('2023-10-15').quarter()") == 4
    assert zen.evaluate_expression("d('2023-01-15').quarter()") == 1
    assert zen.evaluate_expression("d('2023-04-15').quarter()") == 2
    assert zen.evaluate_expression("d('2023-07-15').quarter()") == 3
    # assert zen.evaluate_expression("d('2023-10-15').timestamp()") == 1697328000000   # 这个默认是系统的时区
    # assert zen.evaluate_expression("d('2023-10-15','UTC').timestamp()") == 1697328000000
    assert zen.evaluate_expression("d('2023-10-15', 'Europe/Berlin').offsetName()") == 'Europe/Berlin'
    assert zen.evaluate_expression("d('2023-10-15', 'America/Los_Angeles').offsetName()") == 'America/Los_Angeles'
    assert zen.evaluate_expression("d('2023-10-15').isLeapYear()") == False
    assert zen.evaluate_expression("d('2024-10-15').isLeapYear()") == True
    assert zen.evaluate_expression("d('2000-10-15').isLeapYear()") == True
    assert zen.evaluate_expression("d('1900-10-15').isLeapYear()") == False

    # Date setters
    assert zen.evaluate_expression("d('2023-10-15','UTC').set('year', 2024)") == '2024-10-15T00:00:00Z'    # 默认是当前系统时区 或者设为utc
    assert zen.evaluate_expression("d('2023-10-15','UTC').set('month', 5)") == '2023-05-15T00:00:00Z'
    assert zen.evaluate_expression("d('2023-10-15','UTC').set('day', 20)") == '2023-10-20T00:00:00Z'
    assert zen.evaluate_expression("d('2023-10-15T10:30:00Z').set('hour', 15)") == '2023-10-15T15:30:00+08:00'
    assert zen.evaluate_expression("d('2023-10-15T10:30:00Z').set('minute', 45)") == '2023-10-15T10:45:00+08:00'
    assert zen.evaluate_expression("d('2023-10-15T10:30:00Z').set('second', 30)") == '2023-10-15T10:30:30+08:00'

    # Special date checks
    assert zen.evaluate_expression("d().isSame(d(), 'day')") == True
    assert zen.evaluate_expression("d('2023-10-15').startOf('day')") == '2023-10-15T00:00:00+08:00'
    assert zen.evaluate_expression("d('2023-10-15T10:30:45Z').startOf('hour')") == '2023-10-15T10:00:00+08:00'   # 这个要注意
    assert zen.evaluate_expression("d('2023-10-15').endOf('day')") == '2023-10-15T23:59:59+08:00'
    assert zen.evaluate_expression("d('2023-10-15T10:30:45Z').endOf('hour')") == '2023-10-15T10:59:59+08:00'
    assert zen.evaluate_expression("d('2023-10-15').startOf('month')") == '2023-10-01T00:00:00+08:00'
    assert zen.evaluate_expression("d('2023-10-15').endOf('month')") == '2023-10-31T23:59:59+08:00'
    assert zen.evaluate_expression("d('2023-10-15').startOf('year')") == '2023-01-01T00:00:00+08:00'
    assert zen.evaluate_expression("d('2023-10-15').endOf('year')") == '2023-12-31T23:59:59+08:00'
    assert zen.evaluate_expression("d('2023-10-15').startOf('week')") == '2023-10-09T00:00:00+08:00'
    assert zen.evaluate_expression("d('2023-10-15').endOf('week')") == '2023-10-15T23:59:59+08:00'
    assert zen.evaluate_expression("d('2023-10-15').startOf('quarter')") == '2023-10-01T00:00:00+08:00'
    assert zen.evaluate_expression("d('2023-10-15').endOf('quarter')") == '2023-12-31T23:59:59+08:00'
    assert zen.evaluate_expression("d('2023-03-15').startOf('quarter')") == '2023-01-01T00:00:00+08:00'
    assert zen.evaluate_expression("d('2023-03-15').endOf('quarter')") == '2023-03-31T23:59:59+08:00'
    assert zen.evaluate_expression("d('2023-06-15').startOf('quarter')") == '2023-04-01T00:00:00+08:00'
    assert zen.evaluate_expression("d('2023-06-15').endOf('quarter')") == '2023-06-30T23:59:59+08:00'
    assert zen.evaluate_expression("d('2023-09-15').startOf('quarter')") == '2023-07-01T00:00:00+08:00'
    assert zen.evaluate_expression("d('2023-09-15').endOf('quarter')") == '2023-09-30T23:59:59+08:00'

    # Timezone operations
    assert zen.evaluate_expression("d('2023-10-15T00:00:00Z').tz('America/New_York')") == '2023-10-14T12:00:00-04:00'
    assert zen.evaluate_expression("d('2023-10-15T00:00:00Z').tz('Europe/London')") == '2023-10-14T17:00:00+01:00'
    assert zen.evaluate_expression("d('2023-10-15T00:00:00Z').tz('Asia/Shanghai')") == '2023-10-15T00:00:00+08:00'
    assert zen.evaluate_expression("d('2023-10-15T00:00:00+00:00').tz('UTC')") == '2023-10-14T16:00:00Z'
    assert zen.evaluate_expression("d('2023-10-15T12:00:00+02:00', 'Etc/GMT-2').tz('UTC')") == '2023-10-15T10:00:00Z'

    # Relative date methods
    assert zen.evaluate_expression("d().sub(1, 'd').isYesterday()") == True
    assert zen.evaluate_expression("d().add(1, 'd').isTomorrow()") == True
    assert zen.evaluate_expression("d('2023-10-15').isToday()") == False

    # Diff
    assert zen.evaluate_expression("d('2023-10-15').diff(d('2023-10-10'), 'day')") == 5
    assert zen.evaluate_expression("d('2023-10-15T10:00:00Z').diff(d('2023-10-15T08:30:00Z'), 'hour')") == 1
    assert zen.evaluate_expression("d('2023-10-15').diff(d('2023-09-15'), 'month')") == 1
    assert zen.evaluate_expression("d('2023-12-31').diff(d('2023-01-01'), 'year')") == 0
    assert zen.evaluate_expression("d('2023-12-31').diff(d('2022-01-01'), 'year')") == 1
    assert zen.evaluate_expression("d('2023-10-10').diff(d('2023-10-15'), 'day')") == -5
    assert zen.evaluate_expression("d('2023-09-15').diff(d('2023-10-15'), 'month')") == -1
    assert zen.evaluate_expression("d('2022-01-01').diff(d('2023-12-31'), 'year')") == -1
    assert zen.evaluate_expression("d('2023-10-15T10:30:00Z').diff(d('2023-10-15T10:00:00Z'), 'minute')") == 30
    assert zen.evaluate_expression("d('2023-10-15T10:00:30Z').diff(d('2023-10-15T10:00:00Z'), 'second')") == 30
    assert zen.evaluate_expression("d('2023-10-22').diff(d('2023-10-15'), 'week')") == 1
    assert zen.evaluate_expression("d('2023-10-29').diff(d('2023-10-15'), 'week')") == 2
    assert zen.evaluate_expression("d('2023-07-01').diff(d('2023-01-01'), 'quarter')") == 2
    assert zen.evaluate_expression("d('2023-12-31').diff(d('2023-01-01'), 'quarter')") == 3
    assert zen.evaluate_expression("d('2023-10-15').diff(d('2023-10-15'), 'day')") == 0
    assert zen.evaluate_expression("d('2023-10-15').diff(d('2023-10-15'), 'month')") == 0
    assert zen.evaluate_expression("d('2023-10-15').diff(d('2023-10-15'), 'year')") == 0
    assert zen.evaluate_expression("d('2023-03-31').diff(d('2023-02-28'), 'month')") == 1
    assert zen.evaluate_expression("d('2024-02-29').diff(d('2024-01-31'), 'month')") == 1
    assert zen.evaluate_expression("d('2023-05-31').diff(d('2023-04-30'), 'month')") == 1
    assert zen.evaluate_expression("d('2023-03-31').diff(d('2023-02-28'), 'month')") == 1
    assert zen.evaluate_expression("d('2023-02-28').diff(d('2023-01-31'), 'month')") == 1
    assert zen.evaluate_expression("d('2023-04-30').diff(d('2023-03-31'), 'month')") == 1
    assert zen.evaluate_expression("d('2024-02-29').diff(d('2023-02-28'), 'year')") == 1
    assert zen.evaluate_expression("d('2023-12-31').diff(d('2023-01-01'), 'year')") == 0
    assert zen.evaluate_expression("d('2024-01-01').diff(d('2023-01-01'), 'year')") == 1
    assert zen.evaluate_expression("d('2025-01-01').diff(d('2020-01-01'), 'year')") == 5
    assert zen.evaluate_expression("d('2023-12-01').diff(d('2023-01-01'), 'month')") == 11
    assert zen.evaluate_expression("d('2023-12-31').diff(d('2023-01-01'), 'day')") == 364
    assert zen.evaluate_expression("d('2023-10-15T12:00:00Z').diff(d('2023-10-15T10:30:00Z'), 'hour')") == 1
    assert zen.evaluate_expression("d('2023-10-16T12:00:00Z').diff(d('2023-10-15T18:00:00Z'), 'day')") == 0

    # Formatting
    assert zen.evaluate_expression("d('2023-10-15').format('%Y-%m-%d')") == '2023-10-15'
    assert zen.evaluate_expression("d('2023-10-15').format('%Y/%m/%d')") == '2023/10/15'
    assert zen.evaluate_expression(
        "d('2023-10-15T14:30:45Z').format('%A, %B %d %Y, %H:%M:%S')") == 'Sunday, October 15 2023, 14:30:45'
    assert zen.evaluate_expression("d('2023-10-15T14:30:45Z').format('%a %b %d %H:%M')") == 'Sun Oct 15 14:30'
    assert zen.evaluate_expression("d('2023-10-15T14:30:45Z').format('Day %j of year %Y')") == 'Day 288 of year 2023'
    assert zen.evaluate_expression("d('2023-10-15T14:30:45.123Z').format('%H:%M:%S.%f')") == '14:30:45.123000000'

    # Date validations
    assert zen.evaluate_expression("d(null).isValid()") == False
    assert zen.evaluate_expression("d('foo').isValid()") == False
    assert zen.evaluate_expression("d('2023-13-01').isValid()") == False
    assert zen.evaluate_expression("d('2023-02-30').isValid()") == False

    # Date comparison operators
    assert zen.evaluate_expression("d('2023-10-15') == d('2023-10-15')") == True
    assert zen.evaluate_expression("d('2023-10-15') == d('2023-10-16')") == False
    assert zen.evaluate_expression("d('2023-10-15') != d('2023-10-16')") == True
    assert zen.evaluate_expression("d('2023-10-15') != d('2023-10-15')") == False
    assert zen.evaluate_expression("d('2023-10-15') < d('2023-10-16')") == True
    assert zen.evaluate_expression("d('2023-10-15') < d('2023-10-15')") == False
    assert zen.evaluate_expression("d('2023-10-15') <= d('2023-10-15')") == True
    assert zen.evaluate_expression("d('2023-10-16') <= d('2023-10-15')") == False
    assert zen.evaluate_expression("d('2023-10-16') > d('2023-10-15')") == True
    assert zen.evaluate_expression("d('2023-10-15') > d('2023-10-15')") == False
    assert zen.evaluate_expression("d('2023-10-15') >= d('2023-10-15')") == True
    assert zen.evaluate_expression("d('2023-10-15') >= d('2023-10-16')") == False

    # Date range operators
    assert zen.evaluate_expression("d('2023-10-15') in [d('2023-10-01')..d('2023-10-31')]") == True
    assert zen.evaluate_expression("d('2023-10-01') in [d('2023-10-01')..d('2023-10-31')]") == True
    assert zen.evaluate_expression("d('2023-10-31') in [d('2023-10-01')..d('2023-10-31')]") == True
    assert zen.evaluate_expression("d('2023-09-30') in [d('2023-10-01')..d('2023-10-31')]") == False
    assert zen.evaluate_expression("d('2023-11-01') in [d('2023-10-01')..d('2023-10-31')]") == False
    assert zen.evaluate_expression("d('2023-10-15') in (d('2023-10-01')..d('2023-10-31'))") == True
    assert zen.evaluate_expression("d('2023-10-01') in (d('2023-10-01')..d('2023-10-31'))") == False
    assert zen.evaluate_expression("d('2023-10-31') in (d('2023-10-01')..d('2023-10-31'))") == False
    assert zen.evaluate_expression("d('2023-10-02') in (d('2023-10-01')..d('2023-10-31'))") == True
    assert zen.evaluate_expression("d('2023-10-30') in (d('2023-10-01')..d('2023-10-31'))") == True
    assert zen.evaluate_expression("d('2023-10-15') in [d('2023-10-01')..d('2023-10-31'))") == True
    assert zen.evaluate_expression("d('2023-10-01') in [d('2023-10-01')..d('2023-10-31'))") == True
    assert zen.evaluate_expression("d('2023-10-31') in [d('2023-10-01')..d('2023-10-31'))") == False
    assert zen.evaluate_expression("d('2023-10-30') in [d('2023-10-01')..d('2023-10-31'))") == True
    assert zen.evaluate_expression("d('2023-10-15') in (d('2023-10-01')..d('2023-10-31')]") == True
    assert zen.evaluate_expression("d('2023-10-01') in (d('2023-10-01')..d('2023-10-31')]") == False
    assert zen.evaluate_expression("d('2023-10-31') in (d('2023-10-01')..d('2023-10-31')]") == True
    assert zen.evaluate_expression("d('2023-10-02') in (d('2023-10-01')..d('2023-10-31')]") == True
    assert zen.evaluate_expression("d('2023-10-15') not in [d('2023-10-01')..d('2023-10-31')]") == False
    assert zen.evaluate_expression("d('2023-09-30') not in [d('2023-10-01')..d('2023-10-31')]") == True
    assert zen.evaluate_expression("d('2023-11-01') not in [d('2023-10-01')..d('2023-10-31')]") == True
    assert zen.evaluate_expression("d('2023-10-01') not in (d('2023-10-01')..d('2023-10-31'))") == True
    assert zen.evaluate_expression("d('2023-10-31') not in (d('2023-10-01')..d('2023-10-31'))") == True
    assert zen.evaluate_expression("d('2023-10-15') not in (d('2023-10-01')..d('2023-10-31'))") == False
    assert zen.evaluate_expression("d('2023-10-31') not in [d('2023-10-01')..d('2023-10-31'))") == True
    assert zen.evaluate_expression("d('2023-10-01') not in (d('2023-10-01')..d('2023-10-31')]") == True

    assert zen.evaluate_expression("min([d('2023-01-15'), d('2023-03-20'), d('2023-02-10')])") != '2023-01-15T00:00:00Z'
    assert zen.evaluate_expression("max([d('2023-01-15'), d('2023-03-20'), d('2023-02-10')])") != '2023-03-20T00:00:00Z'
    assert zen.evaluate_expression("min([d('2023-10-15'), d('2023-10-01'), d('2023-10-31')])") != '2023-10-01T00:00:00Z'
    assert zen.evaluate_expression("max([d('2023-10-15'), d('2023-10-01'), d('2023-10-31')])") != '2023-10-31T00:00:00Z'
    assert zen.evaluate_expression(
        "min([d('2023-10-15T10:30:00Z'), d('2023-10-15T08:15:00Z'), d('2023-10-15T14:45:00Z')])") != '2023-10-15T08:15:00Z'
    assert zen.evaluate_expression(
        "max([d('2023-10-15T10:30:00Z'), d('2023-10-15T08:15:00Z'), d('2023-10-15T14:45:00Z')])") != '2023-10-15T14:45:00Z'
    assert zen.evaluate_expression("min([d('2022-12-31'), d('2023-01-01'), d('2024-01-01')])") != '2022-12-31T00:00:00Z'
    assert zen.evaluate_expression("max([d('2022-12-31'), d('2023-01-01'), d('2024-01-01')])") != '2024-01-01T00:00:00Z'
    assert zen.evaluate_expression("min([d('2023-10-15'), d('2023-10-15'), d('2023-10-15')])") != '2023-10-15T00:00:00Z'
    assert zen.evaluate_expression("max([d('2023-10-15'), d('2023-10-15'), d('2023-10-15')])") != '2023-10-15T00:00:00Z'
    assert zen.evaluate_expression("min([d('2023-10-15')])") != '2023-10-15T00:00:00Z'
    assert zen.evaluate_expression("max([d('2023-10-15')])") != '2023-10-15T00:00:00Z'
    assert zen.evaluate_expression(
        "min([d('2023-10-15T10:00:00Z'), d('2023-10-15T12:00:00Z')])") != '2023-10-15T10:00:00Z'
    assert zen.evaluate_expression(
        "max([d('2023-10-15T10:00:00Z'), d('2023-10-15T12:00:00Z')])") != '2023-10-15T12:00:00Z'
    assert zen.evaluate_expression("min([d('1900-01-01'), d('2000-01-01'), d('1950-06-15')])") != '1900-01-01T00:00:00Z'
    assert zen.evaluate_expression("max([d('1900-01-01'), d('2000-01-01'), d('1950-06-15')])") != '2000-01-01T00:00:00Z'
    assert zen.evaluate_expression("min([d('2024-02-29'), d('2023-02-28'), d('2024-03-01')])") != '2023-02-28T00:00:00Z'
    assert zen.evaluate_expression("max([d('2024-02-29'), d('2023-02-28'), d('2024-03-01')])") != '2024-03-01T00:00:00Z'
    assert zen.evaluate_expression("min([d('2023-01-31'), d('2023-02-28'), d('2023-03-31')])") != '2023-01-31T00:00:00Z'
    assert zen.evaluate_expression("max([d('2023-01-31'), d('2023-02-28'), d('2023-03-31')])") != '2023-03-31T00:00:00Z'
    assert zen.evaluate_expression(
        "min([d('2023-10-15').add(1, 'd'), d('2023-10-15').sub(1, 'd'), d('2023-10-15')])") != '2023-10-14T00:00:00Z'
    assert zen.evaluate_expression(
        "max([d('2023-10-15').add(1, 'd'), d('2023-10-15').sub(1, 'd'), d('2023-10-15')])") != '2023-10-16T00:00:00Z'
    assert zen.evaluate_expression("min([d('2023-01-01'), d('2023-12-31')]).isBefore(d('2023-06-01'))") == True
    assert zen.evaluate_expression("max([d('2023-01-01'), d('2023-12-31')]).isAfter(d('2023-06-01'))") == True

    assert zen.evaluate_expression("date('2023-10-15') + duration('1d')") == 1697414400

    assert zen.evaluate_expression("date('2023-10-15 00:00:00+08:00')") == 1697299200
    assert zen.evaluate_expression("date('2023-10-15 08:00:00+08:00')") == 1697328000
    assert zen.evaluate_expression("date('2023-10-15')") == 1697328000

    assert zen.evaluate_expression("year(date('2023-10-15'))") == 2023
    assert zen.evaluate_expression("monthOfYear(date('2023-10-15'))") == 10
    assert zen.evaluate_expression("dayOfMonth(date('2023-10-15'))") == 15
    assert zen.evaluate_expression("dayOfWeek(date('2023-10-15'))") == 7
    assert zen.evaluate_expression("weekOfYear(date('2023-10-15'))") == 41


    assert zen.evaluate_expression("dateString(date('2023-10-15'))") == '2023-10-15 00:00:00'
    assert zen.evaluate_expression("dateString(date('2023-10-15') + duration('12h'))") == '2023-10-15 12:00:00'


def test_special_zen_expression_date_and_time():
    # 特殊 特别注意 涉及到时区的时候，默认是系统的时区

    assert zen.evaluate_expression("d('2023-10-15T10:30:00Z').set('hour', 15)") != '2023-10-15T15:30:00Z'
    assert zen.evaluate_expression("d('2023-10-15T10:30:00Z').set('hour', 15)") == '2023-10-15T15:30:00+08:00'

    assert zen.evaluate_expression("date('2023-10-15 00:00:00+08:00')") == 1697299200
    assert zen.evaluate_expression("date('2023-10-15 08:00:00+08:00')") == 1697328000
    # 系统是东八区
    assert zen.evaluate_expression("date('2023-10-15')") == 1697328000


    # 就算原时区带了Z 表示零时区 带只会取系统的时区
    assert zen.evaluate_expression("d('2023-10-15 14:30')") == '2023-10-15T14:30:00+08:00'  # 当前系统时区为东八区
    assert zen.evaluate_expression("d('2023-10-15T10:30:00Z').set('hour', 15)") == '2023-10-15T15:30:00+08:00'
    assert zen.evaluate_expression("d('2023-10-15T10:30:00Z').set('hour', 15)") != '2023-10-15T15:30:00Z'

    assert zen.evaluate_expression("min([d('2023-01-15'), d('2023-03-20'), d('2023-02-10')])") != '2023-01-15T00:00:00Z'
    assert zen.evaluate_expression("min([d('2023-01-15'), d('2023-03-20'), d('2023-02-10')])") == '2023-01-15T00:00:00+08:00'


def test_zen_expression_array():
    '''
    数组表示其他数据类型的列表（数字、字符串、布尔值列表）
    示例：
    ['a', 'b', 'c'];
    [1, 2, 3]; 
    [true, false];
    '''
    # 数组目前不支持字面量等值判断
    assert not zen.evaluate_expression("[1,2,3]==[1,2,3]") 
    assert not zen.evaluate_expression("['a', 'b', 'c']==['a', 'b', 'c']")

    # 数组成员访问
    assert zen.evaluate_expression("a[1]", { "a": [1, 2, 3, 4]}) == 2

    # 数组关系运算符 in not in
    assert zen.evaluate_expression("1 in a", { "a": [1, 2, 3, 4]}) == True
    assert zen.evaluate_expression("5 not in a", { "a": [1, 2, 3, 4]}) == True

    # 数组范围运算符
    # 闭区间
    assert zen.evaluate_expression("a in [0..2]", {"a": 0}) == True     # 等价 a >= 0 and a <= 2
    assert zen.evaluate_expression("a in (0..2)", {"a": 0}) == False     # 等价 a > 0 and a < 2
    assert zen.evaluate_expression("a in [0..2)", {"a": 1}) == True     # 等价 a >= 0 and a < 2
    assert zen.evaluate_expression("a in (0..2]", {"a": 2}) == True     # 等价 a = 0 and a <= 2

    # 开区间 ]4..6[ 及 )4..6[ 不支持
    assert zen.evaluate_expression("a in )4..6(", {"a": 3}) == True  # 等价 a < 4 or a > 6
    assert zen.evaluate_expression("a in )4..6(", {"a": 7}) == True  # 等价 a < 4 or a > 6
    assert zen.evaluate_expression("a in ]4..6(", {"a": 4}) == True  # 等价 a <= 4 or a > 6
    assert zen.evaluate_expression("a in ]4..6(", {"a": 7}) == True  # 等价 a <= 4 or a > 6

    # 数组函数
    # len 接受一个数字数组并返回所有元素的总和
    assert zen.evaluate_expression("len(a)", {"a": [1, 2, 3]}) == 3

    # sum 接受一个数字数组并返回所有元素的总和
    assert zen.evaluate_expression("sum(a)", {"a": [1, 2, 3]}) == 6
    assert zen.evaluate_expression("sum(a)", {"a": [1.1, 2.53, 3]}) == 6.63

    # avg 接受一个数字数组并返回所有元素的平均值
    assert zen.evaluate_expression("avg(a)", {"a": [1, 2, 3]}) == 2

    # min 接受一个数字数组并返回最小元素
    assert zen.evaluate_expression("min(a)", {"a": [1, 2, 3]}) == 1

    # max 接受一个数字数组并返回最大元素
    assert zen.evaluate_expression("max(a)", {"a": [1, 2, 3]}) == 3

    # zen-engine==0.24.2不支持内置的 mean 方法
    # mean 接受一个数字数组并返回平均值
    # assert zen.evaluate_expression("mean(a)", {"a": [1, 2, 3]}) == 2

    # mode 接受一个数字数组并返回出现最多的元素(众数)
    assert zen.evaluate_expression("mode(a)", {"a": [1, 1, 2, 2, 2, 2, 3, 3]}) == 2

    # contains 接受一个数组和一个搜索参数。如果元素存在于数组中，则返回 true
    assert zen.evaluate_expression("contains(a, b)", {"a": [1, 2, 3], "b": 3}) == True
    assert zen.evaluate_expression("contains(a, b)", {"a": ["a", "b", "c"], "b": "c"}) == True

    # flatten 接受数组并通过单个级别展平参数
    assert zen.evaluate_expression("flatten([1, 'a', ['b', 'c'], [4]])") == [1, "a", "b", "c", 4]
    assert zen.evaluate_expression("flatten([  [1, 2, 3],  [4, 5, 6],])") == [1, 2, 3, 4, 5, 6]

    # 闭包函数
    # all 如果数组的所有元素都满足条件，则返回 true
    assert zen.evaluate_expression("all(a, # in [1..3])", {"a": [1, 2, 3]}) == True
    assert zen.evaluate_expression("all(a, # == 'a')", {"a": ["b", "c"]}) == False

    # some 如果数组中至少有一个元素满足条件，则返回 true
    assert zen.evaluate_expression("some(a, # == 3)", {"a": [1, 2, 3]}) == True
    assert zen.evaluate_expression("some(a, # > 5)", {"a": [1, 2, 3]}) == False

    # none 如果数组中没有元素满足条件，则返回 true
    assert zen.evaluate_expression("none(a, # > 5)", {"a": [1, 2, 3]}) == True
    assert zen.evaluate_expression("none(a, # == 'c')", {"a": ["b", "c"]}) == False

    # filter 返回一个仅包含满足条件的元素的新数组
    assert zen.evaluate_expression("filter(a, # > 1)", {"a": [1, 2, 3]}) == [2, 3]

    # map 返回具有重新映射值的新数组
    assert zen.evaluate_expression("map(a, 'hello ' + #)", {"a": ["world", "user"]}) == ["hello world", "hello user"]
    assert zen.evaluate_expression("map(a, # + 1)", {"a": [1, 2, 3]}) == [2, 3, 4]

    # flatMap 返回具有重新映射值的新数组并将其展平
    assert zen.evaluate_expression("flatMap(a, map(#, # + 1))", {"a": [[1, 2, 3], [4, 5, 6]]}) == [2, 3, 4, 5, 6, 7]

    # count 对数组中的元素按照过滤条件计数
    assert zen.evaluate_expression("count(a, # > 1)", {"a": [1, 2, 3]}) == 2

    # 数组函数复合使用
    assert zen.evaluate_expression("map(filter([1, 2, 3, 4, 5], # > 2), # * 2)") == [6, 8, 10]
    assert zen.evaluate_expression("len(filter([1, 2, 3, 4, 5], # % 2 == 0))")  == 2
    assert zen.evaluate_expression("sum(filter([1, 2, 3, 4, 5], # % 2 == 0))" ) == 6
    assert zen.evaluate_expression("sum(map(filter([1, 2, 3, 4, 5], # > 3), # ^ 2))") == 41

    assert zen.evaluate_expression("some(map([1, 2, 3], # * 2), # > 5)") == True
    assert zen.evaluate_expression("all(map([1, 2, 3], # + 2), # > 2)" ) == True
    assert zen.evaluate_expression("contains(map([1, 2, 3], # * 2), 6)") == True


    
def test_zen_expression_context():
    '''
    Context 是一种特殊的全局数据类型，其结构与节点接收到的 JSON 输入相同。
    示例：
    {  "customer": {    
       "firstName": "John",    
       "lastName": "Doe",    
       "groups": ["admin", "user"],    
       "age": 34  
        }
    }
    '''
    assert zen.evaluate_expression("{a:1}")   == {'a': 1}
    assert zen.evaluate_expression("{'a':1}") == {'a': 1}
    assert zen.evaluate_expression('{"a":1}') == {'a': 1}

    # 注意, object(context) 不支持字面量判断.
    assert not zen.evaluate_expression('{"a":1}=={"a":1}')

    # 属性访问
    assert zen.evaluate_expression("customer.firstName", {"customer": 
      {    
       "firstName": "John",    
       "lastName": "Doe",    
       "groups": ["admin", "user"],    
       "age": 34  
        }
    }) == "John"

    # 运用上述不同数据类型生成表达式
    assert zen.evaluate_expression("customer.firstName + customer.lastName", {"customer":
       {    
       "firstName": "John",    
       "lastName": "Doe",    
       "groups": ["admin", "user"],    
       "age": 34  
        }
    }) == "JohnDoe"

    assert zen.evaluate_expression("customer.age in [30..40]", {"customer":
       {    
       "firstName": "John",    
       "lastName": "Doe",    
       "groups": ["admin", "user"],    
       "age": 34  
        }
    }) == True

    # 多层嵌套
    assert zen.evaluate_expression("user.address.city", {"user":{"address":{"city":"New York"}}}) == "New York"
    assert zen.evaluate_expression("user.contacts[0].phone", {"user":{"contacts":[{"phone":"123-456-7890"}]}}) == "123-456-7890"


def test_zen_expression_ternary_operators():
    assert zen.evaluate_expression("score > 70 ? 'Pass' : 'Fail'", {"score":85}) == 'Pass'
    assert zen.evaluate_expression("score > 90 ? 'A' : score > 80 ? 'B' : score > 70 ? 'C' : 'D'", {"score":85}) == 'B'


import pyparsing as pp

def get_json_parser():
    # Define JSON basic types
    string = pp.QuotedString('"', escChar='\\').setName("string")
    number = pp.pyparsing_common.number().setName("number")
    boolean = (pp.Literal("true") | pp.Literal("false")).setName("boolean")
    null = pp.Literal("null").setName("null")

    # Define JSON array and object structures
    lbracket, rbracket, lbrace, rbrace, colon, comma = map(pp.Suppress, "[]{}:,")

    json_value = pp.Forward().setName("json_value")
    json_member = pp.Group(string + colon + json_value).setName("json_member")
    json_object = pp.Group(pp.Dict(lbrace + pp.Optional(pp.delimitedList(json_member)) + rbrace).setName("json_object"))
    json_array = pp.Group(lbracket + pp.Optional(pp.delimitedList(json_value)) + rbracket).setName("json_array")

    # Define the recursive nature of json_value
    json_value <<= (string | number | boolean | null | json_array | json_object)

    # Define the top-level JSON parser
    json_parser = json_value.setName("json_parser")
    return json_parser


json_parser = get_json_parser()

test_string = '{"name": "Alice", "age": 30, "isStudent": false, "courses": ["Math", "Science"], "address": null}'
parsed_result = json_parser.parseString(test_string)
print(f"{test_string} --> {parsed_result}")


test_string = '[{"name": "Alice", "age": 30}, {"name": "Bob", "age": 25}]'
parsed_result = json_parser.parseString(test_string)
print(f"{test_string} --> {parsed_result}")


test_string = "{'name': 'Alice', 'age': 30}"
parsed_result = json_parser.parseString(test_string)
print(f"{test_string} --> {parsed_result}")



# ; is semicolon
import re

text = r"""foo;;myvar ;;max([5,8,2,11, 7]);;rand(100)+120;;3+4;; 'singel;;quote' ;;"double;;quote" ;;`backquote;; ${bar}`"""

# Regex to find commas not inside double quotes
# This regex works by matching a comma that is NOT followed by an even number of quotes
# until the end of the string.
# It essentially checks if there's an odd number of quotes after the comma,
# implying it's inside a quoted string.
# This pattern is more robust for simple cases without escaped quotes inside strings.
pattern = r""";;(?=(?:[^"'`]*["'`][^"'`]*["'`])*[^"'`]*$)"""

# Find all matches
double_semicolon_outside_quotes = re.findall(pattern, text)
print(double_semicolon_outside_quotes)
print(f"Commas found outside quotes: {double_semicolon_outside_quotes}")

# To split the string by these semicolon:
parts = re.split(pattern, text)
print(f"double semicolon Split parts: {parts}")



#  pytest tests/zen/test_zen.py -sl
#  -s 捕获标准输出
#  -l 显示出错的测试用例中的局部变量
#  -k 指定某一个测试用例执行
#  --setup-show  展示的单元测试 SETUP 和 TEARDOWN 的细节，展示测试依赖的加载和销毁顺序

import asyncio
import json
from pprint import pprint
import pytest
import zen

graphjson = """
{
  "contentType": "application/vnd.gorules.decision",
  "nodes": [
    {
      "type": "inputNode",
      "content": {
        "schema": ""
      },
      "id": "d4c5fdef-96c0-4c5e-8e7b-e1e4f9797251",
      "name": "request",
      "position": {
        "x": 245,
        "y": 355
      }
    },
    {
      "type": "outputNode",
      "content": {
        "schema": ""
      },
      "id": "a2cb6433-140b-4d2c-89f2-0e4f3140dd1a",
      "name": "response",
      "position": {
        "x": 1120,
        "y": 365
      }
    },
    {
      "type": "expressionNode",
      "content": {
        "expressions": [
          {
            "id": "8bf2c094-3115-4608-9416-1c854c68cc89",
            "key": "result",
            "value": "num * 2"
          }
        ],
        "passThrough": false,
        "inputField": null,
        "outputPath": null,
        "executionMode": "single"
      },
      "id": "45bbd693-493b-415d-8ad6-0d57c9a40d3e",
      "name": "expression1",
      "position": {
        "x": 715,
        "y": 400
      }
    }
  ],
  "edges": [
    {
      "id": "84fa9a9a-e90a-4322-be6f-94d3efd84edf",
      "sourceId": "d4c5fdef-96c0-4c5e-8e7b-e1e4f9797251",
      "type": "edge",
      "targetId": "45bbd693-493b-415d-8ad6-0d57c9a40d3e"
    },
    {
      "id": "0cd7ec7d-05d9-4497-abc7-d51b2202ddb4",
      "sourceId": "45bbd693-493b-415d-8ad6-0d57c9a40d3e",
      "type": "edge",
      "targetId": "a2cb6433-140b-4d2c-89f2-0e4f3140dd1a"
    }
  ]
}
"""

graphs = {"expression_multi2.json": graphjson}


# zen engine node type
def test_zen_node_type():
    """
        zen_expression_switch_decisionTable
        这种调用模式可以复用 decision 对象. 所以线上服务推荐使用此种模式
E       RuntimeError: Failed to serialize decision content
E       
E       Caused by:
E           unknown variant `inputNodea`, expected one of `inputNode`, `outputNode`, `functionNode`, 
                            `decisionNode`, `decisionTableNode`, `expressionNode`, `switchNode`, `customNode` at line 12 column 5
    """
    engine = zen.ZenEngine()
    g = json.loads(graphjson)
    nodes = g["nodes"]
    for node in nodes:
        # 修改 inputNode 类型为一个新的类型, 这样执行引擎会报错.
        if node["type"] == "inputNode":
            node["type"] = "inputNode_xxx"
    
    con = json.dumps(g)

    with pytest.raises(RuntimeError) as exc:
        decision = engine.create_decision(con)

    assert "Invalid JSON" in str(exc.value)
    assert "unknown variant `inputNode_xxx`" in str(exc.value)
    node_types = ["inputNode", "outputNode", "functionNode", "decisionNode",
                  "decisionTableNode", "expressionNode", "switchNode", "customNode"]
    assert all([t in str(exc.value) for t in node_types]), f"node type must be {node_types}"





import pyparsing as pp
from pyparsing import QuotedString, Word, alphas, alphanums, nums, pyparsing_common
from pyparsing import infix_notation, OpAssoc, oneOf
from pyparsing import DelimitedList, Forward, Optional, ZeroOrMore, OneOrMore, FollowedBy
# Literal 符号出现在最后的解析中; Suppress 符号不出现在最后的解析中
from pyparsing import Literal, Suppress, Keyword
from pyparsing import Group, Dict, Combine
from pyparsing import originalTextFor, trace_parse_action, ParserElement
# 使用缓存加速递归模式的解析
ParserElement.enablePackrat()


@trace_parse_action
def parse_debug(tokens):
    """
        调试解析到的词法符号 token.
    """
    print("**debug**:", tokens)
    return tokens
    # return ''.join(sorted(set(''.join(tokens))))


def get_zen_object_array_parser(inner_func_originalText=False, debug=False, original_statment=True):
    """
        expressions to parse:
        1. {customer: { firstName: "John", lastName: "Doe" }}  已经完成.
        2. true and true                                       已经完成.
        3.  map([0..3], #);[0, 1, 2, 3]                        已经完成.
            map((0..3], #);[1, 2, 3]
            map([0..3), #);[0, 1, 2]
            map((0..3), #);[1, 2]
            user.contacts[0].phone;{"user":{"contacts":[{"phone":"123-456-7890"}]}};'123-456-7890'  已经完成.
        4. null ?? 'hello'                                        已经完成.
           null ?? null ?? 321
           user.name ?? 'Guest'
           (user.address.city) ?? 'Unknown'
        5. map([{id: 1, name: 'John'}, {id: 2, name: 'Jane'}], #.id)   已经完成. 变量 # 的处理
        6. 'Hello' + ' ' + 'World' + '!'                          已经完成.
        7. `User ${user.name} is ${user.age} years old`;{"user":{"name":"John","age":30}};'User John is 30 years old'  已经完成.
        8. boolean expr ? A : B                                   已经完成.
        9. a+string(xxxx)                                         已经完成.
        10. a.b.c.d                                               已经完成.
        11. a = 5                                                 已经完成.
        12. a=5;b=6                                               未完成解析
    """
    ast_group = originalTextFor if original_statment else Group
    # Define basic elements
    # https://github.com/pyparsing/pyparsing/wiki/Common-Pitfalls-When-Writing-Parsers#identifier--wordalphanums--_-should-be-delimitedlistwordalphas-alphanums--_--combinetrue
    _identifier = Word(alphas + '_#', alphanums + '_#').set_name("identifier")
    # identifier = Word(alphas + "_", alphanums + "_").set_name("identifier")
    identifier = DelimitedList(_identifier, ".", combine=True).set_name("qualified_identifier")

    # Define zen basic types
    double_quotes = QuotedString('"', escChar='\\', unquoteResults=False)
    single_quotes = QuotedString("'", escChar='\\', unquoteResults=False)
    backtick = QuotedString("`", escChar='\\', unquoteResults=False)
    string = (single_quotes | double_quotes | backtick).setName("string")
    number = pyparsing_common.number().setName("number")
    integer = pyparsing_common.integer
    boolean = (Literal("true") | Literal("false")).setName("boolean")
    null = Literal("null").setName("null")
    # Define zen array and object structures
    lbracket, rbracket, lbrace, rbrace, colon, comma = map(Suppress, "[]{}:,")

    zen_value = Forward().setName("zen_value")
    obejct_member = Group((string | identifier) + colon + zen_value).setName("obejct_member")
    zen_object = Dict((lbrace + Optional(DelimitedList(obejct_member)) + rbrace).setName("zen_object"))  # Dict
    zen_array = Group(lbracket + Optional(DelimitedList(zen_value)) + rbracket).setName("zen_array")

    # functions define
    lparen, rparen = Suppress('('), Suppress(')')  # 符号出现在最后的解析中 # lparen, rparen = Suppress('('), Suppress(')')  # 符号不出现在最后的解析中
    args = DelimitedList(ast_group(zen_value))
    function_call = Group(identifier('function_name') + lparen + Optional(args) + rparen)

    ### todo: 还需要解析这种表达式: a=5;b=6 
    assignmentExpr = originalTextFor(identifier.setResultsName("lhs") + Literal("=") + (zen_value).setResultsName("rhs"))

    _range_index_number = Combine(Word(nums) + Optional('.' + Word(nums)) + Optional('e' + Word(nums)))
    # number 这个python 内置的数字表示串会和 .. 冲突, 所以手动指定此处的数字字面量解析
    _range_start_index = (_range_index_number | function_call | identifier).setResultsName("range_start")
    _range_end_index   = (_range_index_number | function_call | identifier).setResultsName("range_end")
    range_expr = originalTextFor(
        Optional((number | function_call | identifier) + Optional(Literal("not")) + Literal("in")) +
        (Literal("[") | Literal("(")) +
        (_range_start_index)  +  
        Literal("..") +
        (_range_end_index) +
        (Literal("]") | Literal(")"))
    )

    atom = range_expr | number | string | boolean | null | function_call | identifier | range_expr
    # Define operator precedence
    _infix_expr = infix_notation(atom, [
        ('-', 1, OpAssoc.RIGHT),
        ('^', 2, OpAssoc.RIGHT),  # Exponentiation
        (oneOf('* / %'), 2, OpAssoc.LEFT),  # Multiplication, Division, Modulo
        (oneOf('+ -'), 2, OpAssoc.LEFT),  # Addition, Subtraction
        (oneOf('> >= < <= == !='), 2, OpAssoc.LEFT),  # Comparison operators
        ('not', 1, OpAssoc.RIGHT),  # Logical NOT
        ('and', 2, OpAssoc.LEFT),  # Logical AND
        ('or', 2, OpAssoc.LEFT),  # Logical OR
        ('??', 2, OpAssoc.LEFT),
        # Define the ternary operator using infix_notation
        # The (("?", ":"), 3, OpAssoc.RIGHT) specifies the ternary operator
        # with '?' and ':' as delimiters, 3 operands, and right-associativity.
        ((Literal("?") , Literal(":")), 3, OpAssoc.RIGHT),
    ])
    infix_expr = ast_group(_infix_expr)

    _slice_index_atom = infix_expr | integer
    _index_expr_ = originalTextFor(
        Suppress("[") + _slice_index_atom + Suppress("]")
    ).setResultsName("index")
    _slice_expr_ = originalTextFor(
        Suppress("[") + Optional(_slice_index_atom) + Literal(":") + Optional(_slice_index_atom) + Suppress("]")
    ).setResultsName("slice")
    ### ^ 代表或, 选择最长的模式去匹配解析
    index_slice_expr = _index_expr_ ^ _slice_expr_

    slice_or_index = originalTextFor(
            (infix_expr | zen_array | string) +  # ## 注意, + 优先级高于 |
            ZeroOrMore(index_slice_expr) + 
            ZeroOrMore(Literal(".") + identifier + ZeroOrMore(index_slice_expr)))
 
    # Define the recursive nature of zen_value
    ### 注意, 需要考虑表达式的顺序,  ^ 代表或, 选择最长的模式去匹配解析, | 表示优先匹配选择, 左边的先匹配.
    # https://github.com/pyparsing/pyparsing/wiki/FAQ---Frequently-Asked-Questions#what-is-the-difference-between-or-and-matchfirst
    zen_value <<= ((function_call ^ infix_expr ^ slice_or_index ^ assignmentExpr) | zen_array | zen_object)  
    # zen_value <<= infix_expr  # 只调试中缀表达式

    # Define the top-level zen parser
    zent_parser = zen_value.setName("zent_parser")
    if debug:
        zent_parser.setDebug()
        zent_parser.set_parse_action(parse_debug)
    return zent_parser

zent_parser = get_zen_object_array_parser(original_statment=True, debug=False)


if __name__ == "__main__":
    ### v3 格式调用
    test_string = "foo(myvar,max([5, 8, 2, 11, 7]),rand(100), 'fccd;;jny', 3+4)" 
    parsed_result = zent_parser.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")

    test_string = "rand(100) >= 0 and rand(100) <= 100" 
    parsed_result = zent_parser.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")


    test_string = "a=3" 
    parsed_result = zent_parser.parseString(test_string)
    print(f"{test_string} --> {parsed_result}")




from contextlib import contextmanager
from contextvars import ContextVar
import argparse
import pymongo


SESSION = ContextVar("session", default=None)


@contextmanager
def transaction(client):
    with client.start_session() as session:
        with session.start_transaction():
            t = SESSION.set(session)
            try:
                yield
            finally:
                SESSION.reset(t)


def insert1(client):
    client.test.txtest1.insert_one({"data": "insert1"}, session=SESSION.get())


def insert2(client):
    client.test.txtest2.insert_one({"data": "insert2"}, session=SESSION.get())


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--url", default="mongodb://localhost:27017")
    args = parser.parse_args()

    client = pymongo.MongoClient(args.url)

    # Create and lear collections, collections must be created outside the transaction
    insert1(client)
    client.test.txtest1.delete_many({})
    insert2(client)
    client.test.txtest2.delete_many({})

    with transaction(client):
        insert1(client)
        insert2(client)

    for doc in client.test.txtest1.find({}):
        print(doc)
    for doc in client.test.txtest2.find({}):
        print(doc)


if __name__ == "__main__":
    main()



#  pytest tests/zen/test_zen.py -sl
#  -s 捕获标准输出
#  -l 显示出错的测试用例中的局部变量
#  -k 指定某一个测试用例执行
#  --setup-show  展示的单元测试 SETUP 和 TEARDOWN 的细节，展示测试依赖的加载和销毁顺序


def test_passing():
    assert (1, 2, 3) == (1, 2, 3)
    # assert {"d": 1, "a":2} == {"d":1}, "两个字典不同"

